/*
 debug.c

 * MacFCEU 0.8a3
 *
 * Copyright notice for this file:
 *  debug.c Copyright (C) 2003 Xodnizel
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "driver.h"
#include "FCEU.h"
#include "debug.h"
#include "general.h"
#include "cheat.h"
#include "cart.h"
#include "GameGenie.h"

#define _RMem16(A)		(ARead[A](A)+(ARead[(A+1)&0xFFFF]((A+1)&0xFFFF)<<8))

/* this was previously in 2A0x.c */
void FCEUI_GetIVectors(uint16 *reset,
	uint16 *irq,
	uint16 *nmi)
{
	int	saveindbg = fceuindbg; //used to just set to 1; 2010.04.04

	fceuindbg = 1;

	if (reset)
		*reset = _RMem16(0xFFFC);
	if (nmi)
		*nmi = _RMem16(0xFFFA);
	if (irq)
		*irq = _RMem16(0xFFFE);

	fceuindbg = saveindbg;
}

void FCEUI_DumpMem(const char *fname,
	uint32 start,
	uint32 end)
{
	FILE		*fp;
	int		olddbg = fceuindbg;

	if (end>0xFFFF)
		end = 0xFFFF;
	if (start>end || (fp=FCEUD_UTF8fopen(fname,"wb"))==0)
		return;

	fceuindbg = 1;

	do
		fputc(ARead[start](start),fp);
	while(++start<=end);

	fceuindbg = olddbg;

	fclose(fp);
}

void FCEUI_LoadMem(const char *fname,
	uint32 start,
	int hl)
{
	FILE		*fp;
	int		t;

	if ((fp=FCEUD_UTF8fopen(fname,"rb"))==0)
		return;

	while((t=fgetc(fp))>=0)
	{
		if (start>0xFFFF)
			break;
		if (!hl)
			BWrite[start](start,t);
		else if (Page[start>>11])
			Page[start>>11][start] = t;
		++start;
	}

	fclose(fp);
}

void FCEUI_MemDump(uint16 a,
	int32 len,
	void (*callb)(uint16 a, uint8 v))
{
	int		olddbg = fceuindbg;

	fceuindbg = 1;

	if (len>0)
		do
			callb(a,ARead[a](a));
		while(++a,--len!=0);

	fceuindbg = olddbg;
}

uint8 FCEUI_MemSafePeek(uint16 A)
{
	int		olddbg = fceuindbg;
	uint8		ret;

	fceuindbg = 1;
	ret = ARead[A](A);
	fceuindbg = olddbg;

	return ret;
}

void FCEUI_MemPoke(uint16 a,
	uint8 v,
	int hl)
{
	if (!hl)
		BWrite[a](a,v);
	else if (Page[a>>11])
		Page[a>>11][a] = v;
}

void DebugGenSetByte(unsigned a,
	unsigned char v)
{
	a&= 0xFFFF;
	FCEUI_MemPoke(a,v,a>=0x800);

	/* make sure it's set in all circumstances; 2010.03.31 */
	if (CheatRPtrs)
	{
		uint8	*d;

		if ((d=CheatRPtrs[a>>10])!=0)
			d[a] = v;
	}
}

static int getAddrName(unsigned a,
	char *n)
{
	static const char	*names2xxx[] =
	{
		"PPU Ctrl #0",			//2000h
		"PPU Ctrl #1",
		"PPU Stat",			//2002h
		"PPU OAM RAM Addr",
		"PPU OAM RAM I/O",		//2004h
		"PPU VRAM Addr Lo",
		"PPU VRAM Addr Hi",		//2006h
		"PPU VRAM I/O"
	};
	static const char	*names4xxx[] =
	{
		"pAPU Rct A Ctrl",		//4000h
		"pAPU Rct A Ramp Ctrl",
		"pAPU Rct A Fine Ctrl",		//4002h
		"pAPU Rct A Coarse Ctrl",
		"pAPU Rct B Ctrl",		//4004h
		"pAPU Rct B Ramp Ctrl",
		"pAPU Rct B Fine Ctrl",		//4006h
		"pAPU Rct B Coarse Ctrl",
		"pAPU Tri Ctrl #0",		//4008h
		"pAPU Tri Ctrl #1",
		"pAPU Tri Freq #0",		//400Ah
		"pAPU Tri Freq #1",
		"pAPU TNs Ctrl #0",		//400Ch
		0, //unimplemented
		"pAPU TNs Freq #0",		//400Eh
		"pAPU TNs Freq #1",
		"pAPU DMC Ctrl",		//4010h
		"pAPU DMC D/A",
		"pAPU DMC Addr",		//4012h
		"pAPU DMC DLen",
		"PPU Sprite DMA",		//4014h
		"pAPU VClock, Chan Ctl",
		"JP/Z A/Strobe, Port Latch",	//4016h
		"JP/Z B/SftClk, Port Latch"
	};
	const char	*s;

	if ((a-=0x2000)&0x80000000U)
		return 0;
	if (a<GET_ARRAY_COUNT(names2xxx))
		s = names2xxx[a];
	else if ((a-=0x4000-0x2000)<GET_ARRAY_COUNT(names4xxx))
		s = names4xxx[a];
	else
		return -1;
	if (n)
		{strcpy(n,s); return 1;}
	return -1;
}

int DebugGetAddrType(uint16 addr,
	char *name)
{
	readfunc	an = ANull;
	writefunc	bn = BNull;
	uint32		a = (uint32)addr;
	int		r = 0;
	/* since the write addresses may be optimized */
	OptFAddr(an,an);
	OptFAddr(bn,bn);

	if (ARead[a]==an && BWrite[a]==bn)
		return -1;	//unmapped
	if (a>=0x8000)
		if (GenieRWWrap)
		{
			if (AReadG && ARead[a]!=AReadG[a-0x8000])
				r|= 4; //is GG read-bypassed
			if (BWriteG && BWrite[a]!=BWriteG[a-0x8000])
				r|= 8; //is GG write-bypassed
		}
	if (a-0x2000<0x4018-0x2000)
	{
		getAddrName(a,name);
		goto end;	//never RAM / PRG
	}
	a>>= 10;
	if (Page[a>>1] && (a&0xE000>>10)!=0x6000>>10) /* added SRAM/WRAM check; 2010.04.08 */
		r|= 1;		//is PRG
	if (PRGIsRAM[a>>1] || CheatRPtrs[a] || !(a&0xE000>>10)) /* ...general RAM... */
		r|= 2;		//is RAM
end:
	return r;
}

/* ---- MacFCEU breakpoints */

typedef struct bpoint/*(8+)*/
{
	uint16		address;
	uint8		original, flags;
	char		*expression; /* only break if parse_parameter(..)!=0 */
} bpoint;

uint8		breakpoint_bitmap[0x10000>>3]; /* all addresses; 8KB */

static bpoint		*breakpoint_list = 0;
static int32		breakpoint_count = 0;

int32 GetBreakPointCount()
{
	return breakpoint_list ? breakpoint_count:0;
}

#define BPGet(b,c,i)		do{ \
	if(breakpoint_count<=0 || i>=breakpoint_count) \
		return -1; \
	if(i<0)	\
		b=breakpoint_list,c=breakpoint_count; \
	else \
		b=i+breakpoint_list,c=1; \
	if(!c) \
		return 0; \
}while(0)

static int32 bpsReset(const bpoint *b,
	int32 c,
	int set)
{
	int32		h = 0;

	if (set)
		do
			if (b->flags&BPOINT_HARD)
				++h;
			else
				DebugGenSetByte(b->address,subsopcode_breakpoint);
		while(++b,--c!=0);
	else
		do
			if (!(b->flags&BPOINT_HARD))
				DebugGenSetByte(b->address,b->original);
		while(++b,--c!=0);

	return h; /* hard count (step through) */
}

int32 DebugBreakPointsReset(int32 index,
	int set)
{
	const bpoint	*b;
	int32		c;

	BPGet(b,c,index);

	return bpsReset(b,c,set);
}

int DebugBreakPointsLoop(int32 index,
	uint8 targetflags,
	int (*callb)(int32 index,
		int32 validcount,
		uint32 value, /* [flags.16] [address.16] */
		const char *expression))
{
	const bpoint	*b;
	int		i, c, vc = 0, t;

	BPGet(b,c,index);

	i = 0;
	do
	{
		if (targetflags!=0 && b->flags!=targetflags)
			continue;
		if ((t=callb(i,vc,(b->flags<<16)|b->address,b->expression))==0)
			continue;
		if (t<0)
			break;
		++vc;
	} while(++b,++i,--c!=0);

	return vc;
}

int32 DebugBreakPointsClear(int32 index)
{
	bpoint		*b;

	if (breakpoint_count<=0)
		return 0;
	if (index<0)
	{
		int32	c;

		DebugBreakPointsReset(-1,0); /*reset all possible softpoints*/
		/*free everything, and clear bits*/
		c = index = breakpoint_count, b = breakpoint_list;
		do
			if (b->expression)
				b->expression = realloc(b->expression,0);
		while(++b,--c!=0);
		breakpoint_list = realloc(breakpoint_list,breakpoint_count=0);
		memset(breakpoint_bitmap,0,sizeof(breakpoint_bitmap));
		return index;
	}
	if (index>=breakpoint_count)
		return -1;
	b = index+breakpoint_list;

	_bitclr(breakpoint_bitmap,b->address);
	/*reset possible softpoint*/
	bpsReset(b,1,0);
	/*free any extra data it may hold*/
	if (b->expression)
		b->expression = realloc(b->expression,0);
	/*remove from list*/
	memmove(b,1+b,(--breakpoint_count-index)*sizeof(bpoint));
	if ((b=realloc(breakpoint_list,breakpoint_count*sizeof(bpoint)))==0
			&& breakpoint_count!=0)
		return ~breakpoint_count;
	/*re-set pointer*/
	breakpoint_list = b;

	return 1;
}

int32 DebugBreakPointGet(uint16 address)
{
	bpoint		*b;
	int32		i = 0, c;

	if ((b=breakpoint_list)!=0 && (c=breakpoint_count)>0)
		do
			if (b->address==address)
				return i;
		while(++b,++i<c);

	return -1;
}

int32 DebugBreakPointSet(uint16 address,
	uint8 flags,
	const char *expression)
{
	bpoint		*b;
	int32		index;
	int		hard = 1;

	if (!(flags&BPOINT_PC))
		return -1;

	if ((index=DebugBreakPointGet(address))>=0)
		b = index+breakpoint_list;
	else
	{ /* add -- technically limited to int32-max breakpoints */
		if ((index=breakpoint_count)<0 || ++index<0
				|| (b=realloc(breakpoint_list,index*sizeof(bpoint)))==0)
			return -1;
		breakpoint_list = b;
		breakpoint_count = index;
		b+= --index;
		b->address = address;
		b->original = FCEUI_MemSafePeek(address);
		b->expression = 0;
	}

	_bitset(breakpoint_bitmap,address);
	b->flags = flags;

	if (expression)
	{
		char	*p;
		int32	z = strlen(expression)+1;

		if ((p=realloc(b->expression,z))!=0)
			memcpy(b->expression=p,expression,z);
	}

	if (!(b->flags&BPOINT_HARD) && address>=0x2000)
	{
		bpsReset(b,1,1);
		hard = FCEUI_MemSafePeek(address)!=subsopcode_breakpoint;
		bpsReset(b,1,0);
	}

	if (hard)
		b->flags|= BPOINT_HARD;

	return index|(hard<<31); /* bit 31 = hard; only -1 means error */
}

/* ---- general disassembly */

#include "x6502.h"
#include "PPU.h"
#include "utility/disasm6502.h"

static long addtracetext(char *buf,
	const X6502 *f,
	uint32 tf)
{
	long		z = strlen(buf);

	if (tf&~tfaddnewline)
	{
		/* complete line tabulation; add white space, if necessary */
		while(z<32)
			buf[z++] = ' ';
		if (f!=NULL)
		{
			if (tf&tfaddCPU_A)
				z+= sprintf(z+buf," A:%02X",
					(uint32)f->A);
			if (tf&tfaddCPU_X)
				z+= sprintf(z+buf," X:%02X",
					(uint32)f->X);
			if (tf&tfaddCPU_Y)
				z+= sprintf(z+buf," Y:%02X",
					(uint32)f->Y);
			if (tf&tfaddCPU_P)
				z+= sprintf(z+buf," P:%02X",
					(uint32)f->P);
			if (tf&tfaddCPU_S)
				z+= sprintf(z+buf," S:%02X",
					(uint32)f->S);
			if (tf&tfaddCPU_IRQlow)
				z+= sprintf(z+buf," IRL:%06X",
					f->IRQlow);
			if (tf&tfaddCPU_CYC)
				z+= sprintf(z+buf," CYC:%2u",
					timestamp);
			if (tf&tfaddPPU_scanline)
				z+= sprintf(z+buf," SL:%2u",
					scanline);
			if (tf&tfaddstack8)
				z+= sprintf(z+buf,f->S&1 ?
					" $1%02X:%02X %02X%02X %02X%02X %02X%02X %02X":
					" $1%02X:%02X%02X %02X%02X %02X%02X %02X%02X",
					f->S,
					RAM[f->S+0x100],
					RAM[(uint8)(f->S+1)+0x100],
					RAM[(uint8)(f->S+2)+0x100],
					RAM[(uint8)(f->S+3)+0x100],
					RAM[(uint8)(f->S+4)+0x100],
					RAM[(uint8)(f->S+5)+0x100],
					RAM[(uint8)(f->S+6)+0x100],
					RAM[(uint8)(f->S+7)+0x100]);
		}
	}
	if (tf&tfaddnewline)
		buf[z++] = '\n';
	buf[z] = 0;

	return z;
}

static int messagePrinter(const char *s,
	int i)
{
(void)i;
	FCEUD_Message(s);

	return 1;
}

uint16 DebugDisassemble65xx(const X6502 *f,
	uint16 pc,
	int c,
	int (*routine)(const char*, int i),
	uint32 traceflags)
{
	int		saveindbg = fceuindbg;

	char		buf[192]; /* I count 131 maximum, but just in case */
	int		i = 0;
	int		fpc = f ? f->PC:-1;

	if (!routine)
		routine = messagePrinter;
	if (c<0)
		f = 0, c =~c;
	if (!c) /* zero: one line, no '>' */
		fpc =-1, c = 1;

	fceuindbg = 1;

	do
	{
		buf[0] = pc==fpc ? '>':' ';
		buf[1] = _bittst(breakpoint_bitmap,pc) ? '*':' ';
		sprintf(2+buf,"%04X  ",
			pc);
		pc = X6502_Disassemble(fpc<0 ? 0:f,ARead,pc,0xFFFF,0,8+buf);
		addtracetext(buf,f,traceflags);
		if (!routine(buf,i))
			f = 0; /* to stop printing info based on present circumstances */
	} while(++i,--c!=0);

	fceuindbg = saveindbg;

	return pc;
}
