/*
 input/Zapper.c

 * MacFCEU 0.6.1
 *
 * Copyright notice for this file:
 *  zapper.c Copyright (C) 2002 Xodnizel
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <string.h>
#include <stdlib.h>

#include "share.h"
#include "../palette.h"
#include "../git.h"
#include "Zapper.h"

ZAPPER	InputZapperData[2];

static void FP_FASTAPASS(3) ZapperFrapper(int w,
	uint8 *bg,
	uint8 *spr,
	uint32 linets,
	int final)
{
 int xs,xe;
 int zx,zy;

 if(!bg) // New line, so reset stuff.
 {
  InputZapperData[w].zappo=0;
  return;
 }
 xs=InputZapperData[w].zappo;
 xe=final;

 zx=InputZapperData[w].mzx;
 zy=InputZapperData[w].mzy;

 if(xe>256) xe=256;

 if(scanline>=(zy-4) && scanline<=(zy+4))
 {
  while(xs<xe)
  {
    uint8 a1,a2;
    uint32 sum;
    if(xs<=(zx+4) && xs>=(zx-4))
    {
     a1=bg[xs];
     if(spr)
     {
      a2=spr[xs];
 
      if(!(a2&0x80))
       if(!(a2&0x40) || (a1&64))
        a1=a2;
     }
     a1&=63;

     sum=palo[a1].r+palo[a1].g+palo[a1].b;
     if(sum>=100*3)
     {
      uint64 t;

    //InputZapperData[w].zaphit=((uint64)linets+(xs+16)*(PAL?15:16))/48+timestampbase; 
	(void)mulu64u32(t,xs+16,PAL?15:16);
	(void)uint64addu32(t,linets);
	(void)uint64divu32(t,48);
	(void)uint64add(t,timestampbase);
	InputZapperData[w].zaphit=t;
      goto endo;
     }
    }   
   xs++;
  }
 }
endo:
 InputZapperData[w].zappo=final;
}      

static INLINE int CheckColor(int w)
{
	uint64		t = InputZapperData[w].zaphit, u = timestampbase;

	FCEUPPU_LineUpdate();
/*if((ZD[w].zaphit+100)>=(timestampbase+timestamp) && !(ZD[w].mzb&2)) return(0); */
	(void)uint64addu32(t,100);
	(void)uint64addu32(u,timestamp);
	return int64ge(t,u) && (InputZapperData[w].mzb&2)==0 ? 0:1; //2011.02.05
}

static uint8 FP_FASTAPASS(1) ReadZapperVS(int w)
{
	uint8	ret = 0;

	if (InputZapperData[w].zap_readbit==4)
		ret = 1;

	if (InputZapperData[w].zap_readbit==7)
		if (InputZapperData[w].bogo)
			ret|= 1;
	if (InputZapperData[w].zap_readbit==6)
		ret|= CheckColor(w);
	if (!fceuindbg)
		++InputZapperData[w].zap_readbit; 
	return ret;
}

static void FP_FASTAPASS(1) StrobeZapperVS(int w)
{
	InputZapperData[w].zap_readbit = 0;
}

static uint8 FP_FASTAPASS(1) ReadZapper(int w)
{
	uint8		ret = 0;

	if (InputZapperData[w].bogo)
		ret|= 0x10;
	ret|= CheckColor(w)<<3; //if...|=8
	return ret;
}

static void FASTAPASS(3) DrawZapper(int w, uint8 *buf, int arg)
{
	if(arg)
		FCEU_DrawGunSight(buf,256,InputZapperData[w].mzx,InputZapperData[w].mzy);
}

static void FP_FASTAPASS(3) UpdateZapper(int w, void *data, int arg)
{
	uint32	*ptr = (uint32*)data;
(void)arg;
//FCEU_DispMessage("%3d:%3d",ZD[w].mzx,ZD[w].mzy);
	if (InputZapperData[w].bogo)
		--InputZapperData[w].bogo;
	if (ptr[2]&3 && (InputZapperData[w].mzb&3)==0)
		InputZapperData[w].bogo = 5;

	InputZapperData[w].mzx = ptr[0];
	InputZapperData[w].mzy = ptr[1];
	InputZapperData[w].mzb = ptr[2];
}

static INPUTC ZAPC =
	{ReadZapper,0,0,UpdateZapper,ZapperFrapper,DrawZapper};
static INPUTC ZAPVSC =
	{ReadZapperVS,0,StrobeZapperVS,UpdateZapper,ZapperFrapper,DrawZapper};

INPUTC *FCEU_InitZapper(int w)
{
	memset(&InputZapperData[w],0,sizeof(ZAPPER));
	if (FCEUGameInfo->type==GIT_VSUNI)
		return &ZAPVSC;
	else
		return &ZAPC;
}
