/*
 macos/listwindow.c

  Created: 2013.05.06 (orig. from messagelog.c)
  Revised: 2013.05.20
*/
/*
 TODO: use GetDblTime for double-click indication.
*/

#include <stdlib.h>
#include <string.h>

#include <Events.h>
#include <Controls.h>
#include <ControlDefinitions.h>
#include <Scrap.h>

#include "listwindow.h"

/* ---- constants */

/* 256 bytes buffered; a Classic DialogRecord is 170 bytes */
#define LISTWINDOW_SELSTART(w)		(*(int32*)(200+(Ptr)(w)))
#define LISTWINDOW_SELEND(w)		(*(int32*)(204+(Ptr)(w)))

#define LISTWINDOW_TLINELIST(w)		(*(int32**)(208+(Ptr)(w)))
#define LISTWINDOW_TFIRSTLN(w)		(*(int32*)(212+(Ptr)(w))) /* where to clip when appending */
#define LISTWINDOW_TLENGTH(w)		(*(int32*)(216+(Ptr)(w)))
#define LISTWINDOW_TEXT(w)		(*(char**)(220+(Ptr)(w)))
#define LISTWINDOW_MAXLCOUNT(w)		(*(uint32*)(224+(Ptr)(w)))
#define LISTWINDOW_INDENT(w)		(*(Point*)(228+(Ptr)(w)))
#define LISTWINDOW_MAXLSIZE(w)		(*(int16*)(232+(Ptr)(w)))

#define LW_FULL_UPDATE			0x8000
#define LW_MODAL			0x80
#define LW_SLECTABLE			0x1
#define LW_MULTIPLESEL			0x2

#define LISTWINDOW_FLAG(w)		(*(int16*)(234+(Ptr)(w)))
#define LISTWINDOW_LHGT(w)		(*(int16*)(236+(Ptr)(w)))
#define LISTWINDOW_BOTTOMPAD(w)		(*(int16*)(238+(Ptr)(w)))

#define LISTWINDOW_LCOUNT(w)		(*(int32*)(240+(Ptr)(w)))
#define LISTWINDOW_VSCROLL(w)		(*(ControlRef*)(244+(Ptr)(w)))
#define LISTWINDOW_HSCROLL(w)		(*(ControlRef*)(248+(Ptr)(w)))

#define SCROLL_DOWN_DELAY		20 /* in 60ths ticks - was 30 */

#define DEFAULT_INDENT_H		6
#define DEFAULT_INDENT_V		0

/* ---- macros */

#define ShowHideControl(c,x)		do{if(x) ShowControl(c); else HideControl(c);}while(0);

#if !TARGET_API_MAC_CARBON
 /* this is actually an Appearance API */
 #define IsControlVisible(h)		(0+(**(h)).contrlVis)
#endif

/* ---- globals */

static ControlActionUPP		lwctlupp = NULL;

/* ---- code */

/* appending text to a clipped end requires that we SCAN IN REVERSE */
/* returns the earliest list index */
int32 rcalclines(int32 *list,
	int32 *count, /* required */
	const char *text,
	const char *e)
{
	int32		*l;
	const char	*s;
	long		t = 0; /* was 1 - the append func will increment when appropriate */
	uint8		p = 0xFF;

	if (!count || text>=(s=e))
		return -1;
	if ((l=list)!=NULL)
		l+= *count;
	do
	{
		uint8		c = *--s;

		if ((c-=10)<=13-10) /* any EOL */
		{ /* p==0 means the following char is a linefeed */
			if (!p && c==(uint8)(13-10)) /* DOS line ending (skip) */
				{p = 0xFF; continue;}
			++t;
			if (l)
				if (--l<=list)
					{++s, ++l; break;}
				else
					*l = 1+s-text;
		}
		p = c;
	} while(s>text);

	if (l && --l>=list)
		*l = s-text;
	*count = t;

	return l-list;
}

int listwindow_append(WindowRef window,
	const char *str,
	int z,
	Boolean update)
{
	char		*t;
	int32		max, total;

	if (!window)
		return -1;
	if ((total=LISTWINDOW_TLENGTH(window))<0)
		total = 0;
	max = LISTWINDOW_MAXLCOUNT(window);
	if (!LISTWINDOW_TLINELIST(window))
		if ((LISTWINDOW_TLINELIST(window)=calloc(1,max*sizeof(int32)))==NULL)
			return -2;
	if ((t=LISTWINDOW_TEXT(window))!=NULL && max
			&& LISTWINDOW_LCOUNT(window)>=max)
	{ /* clip */
		int		d;

		if ((d=LISTWINDOW_TFIRSTLN(window))<0)
			return -1;
		LISTWINDOW_FLAG(window)|= LW_FULL_UPDATE;
		//if (LISTWINDOW_TLINELIST(window)[d]<0)
		//	LISTWINDOW_TFIRSTLN(window) = ++d;
		if ((d=LISTWINDOW_TLINELIST(window)[d])>=0)
			if ((total-=d)<0)
				total = 0;
			else
				memmove(t,t+d,total);
	}
	LISTWINDOW_TLENGTH(window) = total;

	if (z<=0)
		return 0;
	if ((t=realloc(t,total+z))==NULL)
		return -2;
	LISTWINDOW_TEXT(window) = t;
	LISTWINDOW_TLENGTH(window) = total+z;

	memcpy(total+t,str,z);

	if (!max)
	{ /* no clipping: gather on what's added instead */
		int32		*l;
		const char	*n = t+z, *e = t+total;
		int32		c = 0;
		/* count lines - account for a newly formed CR-LF sequence too */
		/* apparently, MPW is lax on this... (graphical errors pasting disparate eol types) */
		if (total>0 && t[total-1]==13 && *str==10)
			++n; /* skip to discount */
		rcalclines(NULL,&c,n,e);
		/* reallocate */
		max = LISTWINDOW_LCOUNT(window)+c;
		if ((l=realloc(LISTWINDOW_TLINELIST(window),max*sizeof(int32)))==NULL)
			return -2;
		/* dealing with offsets that don't change - only add new ones (faster) */
		rcalclines((LISTWINDOW_TLINELIST(window)=l)+LISTWINDOW_LCOUNT(window),&c,n,e);
		/* now we can update the line count */
		LISTWINDOW_LCOUNT(window) = max;
		LISTWINDOW_TFIRSTLN(window) = 0;
	} else
	{ /* initialize fixed maximum for rcalclines (to do the whole thing) */
		LISTWINDOW_LCOUNT(window) = max;
		LISTWINDOW_TFIRSTLN(window) = rcalclines(LISTWINDOW_TLINELIST(window),
			&LISTWINDOW_LCOUNT(window),t,t+LISTWINDOW_TLENGTH(window));
		if (LISTWINDOW_LCOUNT(window)<max)
			++LISTWINDOW_LCOUNT(window);
	}
	if (update)
	{
		listwindow_updatescrolls(window,false);
		listwindow_update(window,!!(LISTWINDOW_FLAG(window)&LW_FULL_UPDATE));
		LISTWINDOW_FLAG(window)&=~LW_FULL_UPDATE;
	}

	return 1;
}

static int listwindow_free(WindowRef window) /* should be valid */
{
	int		r = 0;

	if (LISTWINDOW_TEXT(window))
		LISTWINDOW_TEXT(window) = realloc(LISTWINDOW_TEXT(window),0), r = 1;
	LISTWINDOW_TLENGTH(window) = LISTWINDOW_LCOUNT(window) = 0;
	if (LISTWINDOW_TLINELIST(window))
		LISTWINDOW_TLINELIST(window) = realloc(LISTWINDOW_TLINELIST(window),0);
	return r;
}

int listwindow_clear(WindowRef window)
{
	if (!window)
		return -1;
	listwindow_free(window);
	listwindow_updatescrolls(window,false);

	return listwindow_update(window,true);
}

int listwindow_select(WindowRef window,
	int32 start,
	int32 end)
{
	int		c = 0;

	if (!window)
		return -1;
	if (start<0)
		start = 0;
	if (end>=LISTWINDOW_LCOUNT(window))
		end = LISTWINDOW_LCOUNT(window)-1;
	if (LISTWINDOW_SELSTART(window)!=start)
		LISTWINDOW_SELSTART(window) = start, c = 1;
	if (LISTWINDOW_SELEND(window)!=end)
		LISTWINDOW_SELEND(window) = end, ++c;
	return c;
}

int listwindow_getsel(WindowRef window,
	int32 *start,
	int32 *end)
{
	int		s, e;

	if (!window)
		return -1;
	e = LISTWINDOW_MAXLCOUNT(window)-1;
	s = PIN(LISTWINDOW_SELSTART(window),0,e);
	s = PIN(LISTWINDOW_SELEND(window),0,e);
	if (start)
		*start = s;
	if (end)
		*end = e;
	return LISTWINDOW_SELEND(window)>=0 ? e-s+1:0;
}

int listwindow_getlheight(WindowRef window)
{
	int		v;

	if ((v=LISTWINDOW_LHGT(window))==0)
		v = 1;
	return v;
}

int listwindow_getscrolls(WindowRef window,
	int *h,
	int *v)
{
	ControlRef	control;
	int		c;

	if (!window)
		return -1;
	c = 0;
	if ((control=LISTWINDOW_VSCROLL(window))!=NULL)
	{
		++c;
		if (v)
			*v = GetControlValue(control);
	}
	if ((control=LISTWINDOW_HSCROLL(window))!=NULL)
	{
		++c;
		if (h)
			*h = GetControlValue(control);
	}
	return c;
}

int listwindow_drawtext(WindowRef window,
	int c)
{
	const int32	*a, *e;
	const char	*t;
	int		i, h, x, y, sl, el, lheight, mxlsz;
	Boolean		hl;

	if (!window)
		return -1;
	if ((t=LISTWINDOW_TEXT(window))==NULL
			|| (a=LISTWINDOW_TLINELIST(window))==NULL
			|| (i=LISTWINDOW_TFIRSTLN(window))<0)
		return 0;
	if (LISTWINDOW_MAXLCOUNT(window))
		a+= i;
	e = a-1+LISTWINDOW_LCOUNT(window);
	i = x = 0, mxlsz = LISTWINDOW_MAXLSIZE(window)+1;

	listwindow_getscrolls(window,&x,&i);
	x*=-6;
	x+= LISTWINDOW_INDENT(window).h;
	y = LISTWINDOW_INDENT(window).v+(lheight=listwindow_getlheight(window));

	/* completion to account for scrolling */
	if (i>0)
		--i, y-= lheight, ++c;
	if ((unsigned)(h=LISTWINDOW_LCOUNT(window)-i)>c)
		h = c;
	a+= i;
	hl = false, sl = LISTWINDOW_SELSTART(window), el = LISTWINDOW_SELEND(window);
	TextMode(srcCopy);
	do
	{
		const char	*s;
		long		z;
		Boolean		final;

		MoveTo(x,y);
		if (a>e)
			break;
		final = a>=e;
		if ((z=(final ? LISTWINDOW_TLENGTH(window):a[1])-*a)<=0)
			continue;
		s = t+*a;
		if (s[z-1]=='\012') /* dont draw LF; 2009.06.27 */
			--z;
		if (i>=sl && i<=el)
		{
			if (!hl)
				{hl = true; TextMode(notSrcCopy);}
		} else if (hl)
			{hl = false; TextMode(srcCopy);}
		DrawText(s,0,MIN(z,mxlsz));
		if (!final)
			DrawChar(' ');
	} while(++a,y+=lheight,++i,--h);
	TextMode(srcCopy);

	return 1;
}

static void getlistbounds(WindowRef window,
	Rect *bounds)
{
	//if (!window || !bounds)
	//	return;
	getwindowbounds(window,bounds);
	if (LISTWINDOW_VSCROLL(window))
		bounds->bottom-= 16;
	if (LISTWINDOW_HSCROLL(window))
		bounds->right-= 16;
}

static void listwindow_scrollbox(WindowRef window, /* should be valid */
	short sh,
	short sv)
{
	Rect		newfocus;
	RgnHandle	region;
	GrafPtr		oldport;

	GetPort(&oldport);
	setportwindow(window);

	getlistbounds(window,&newfocus);
	region = NewRgn();
	ScrollRect(&newfocus,sh,sv,region);

	if (region) /* update what ScrollRect has given us (the remainder that needs rendering) */
		SetClip(region);
	listwindow_drawtext(window,(newfocus.bottom-newfocus.top)/listwindow_getlheight(window));
	if (region)
		DisposeRgn(region);
	/* reset clip (anything large enough will do) */
	*(int32*)&newfocus.top = 0;
	*(int32*)&newfocus.bottom = 0x40004000;
	ClipRect(&newfocus);

	MacSetPort(oldport);
}

int listwindow_scroll(WindowRef window,
	int hcount,
	int vcount)
{
	ControlRef	control;
	int		hsep = IMIN(int), vsep = IMIN(int);
	int		c = 0, vshift = vcount&(cmdKey<<16) ? 11:0;

	if (!window)
		return -1;
	/* values above 0xFFFF are page move indicators */
	if (hcount>0xFFFF)
	{
		if ((hcount&=0xFF)==0)
		{ /* command page -> vertical to horizontal scroll */
			if ((hcount=vcount&0xFF)>kPageDownCharCode || !vshift)
				{hcount = 0; goto skip;}
			vcount = 0;
		}

		switch(hcount)
		{
		case kPageUpCharCode: hcount =-8; break;
		case kPageDownCharCode: hcount = 8; break;
		case kUpArrowCharCode: hcount =-1; break;
		case kDownArrowCharCode: hcount = 1; break;
		default: hcount = 0;
		}
	}
	if (vcount>0xFFFF)
		switch(vcount&0xFF)
		{
		case kPageUpCharCode: vcount =-8<<vshift; break;
		case kPageDownCharCode: vcount = 8<<vshift; break;
		case kUpArrowCharCode: vcount =-1<<vshift; break;
		case kDownArrowCharCode: vcount = 1<<vshift; break;
		default: vcount = 0;
		}
skip:
	if ((control=LISTWINDOW_VSCROLL(window))!=NULL)
	{
		++c, vsep = GetControlValue(control);
		if ((vcount+=vsep)<-32768)
			vcount =-32768;
		else if (vcount>32767)
			vcount = 32767;
		SetControlValue(control,vcount);
		vsep-= GetControlValue(control);
		vsep*= listwindow_getlheight(window);
	}
	if ((control=LISTWINDOW_HSCROLL(window))!=NULL)
	{
		++c, hsep = GetControlValue(control);
		if ((hcount+=hsep)<-32768)
			hcount =-32768;
		else if (hcount>32767)
			hcount = 32767;
		SetControlValue(control,hcount);
		hsep-= GetControlValue(control);
		hsep*= 6;
	}

	listwindow_scrollbox(window,hsep,vsep);
	LISTWINDOW_FLAG(window)|= LW_FULL_UPDATE;

	return c;
}

int listwindow_updatescrolls(WindowRef window,
	Boolean doBounds)
{
	Rect		bounds;
	ControlRef	control;
	int		value, vertc = 64, width, height;

	if (!window)
		return -1;
	getwindowbounds(window,&bounds);
	height = bounds.bottom-bounds.top-LISTWINDOW_BOTTOMPAD(window);
	width = bounds.right-bounds.left;
	if ((control=LISTWINDOW_VSCROLL(window))!=NULL)
	{ /* vertical */
		int	lheight = listwindow_getlheight(window);

		if (doBounds)
		{
			bounds.top =-1;
			bounds.left = width-15;
			bounds.bottom = height-14;
			bounds.right = width+1;
			SetControlBounds(control,&bounds);
		} else
			GetControlBounds(control,&bounds);
		if ((vertc=LISTWINDOW_LCOUNT(window)-(bounds.bottom-bounds.top-15)/lheight-1)<=0)
			SetControlMaximum(control,0);
		else
		{ /* set maximum, update scroll position if at maximum */
			value = GetControlMaximum(control);

			SetControlMaximum(control,vertc);

			if (GetControlValue(control)==value) /*value>0 && */
			{
				SetControlValue(control,vertc);
				value-= vertc;
				listwindow_scrollbox(window,0,value*lheight);
			}
		}
	}
	if ((control=LISTWINDOW_HSCROLL(window))!=NULL)
	{ /* horizontal */
		if (doBounds)
		{
			bounds.top = height-15;
			bounds.left =-1;
			bounds.bottom = height+1;
			bounds.right = width-14;
			SetControlBounds(control,&bounds);
		} else
			GetControlBounds(control,&bounds);
		value = LISTWINDOW_MAXLSIZE(window)-(bounds.right+bounds.left/6)+6;
		SetControlMaximum(control,value<0 || !LISTWINDOW_TEXT(window) ? 0:value);
	}

	return vertc;
}

static pascal void mycontrolactionproc(ControlRef control,
	ControlPartCode part)
{
	WindowRef	window;
	unsigned long	ticks = TickCount();
	unsigned long	reference = GetControlReference(control);
	SInt16		value;

	if (ticks<reference)
		return;
	value = GetControlValue(control);
	switch(part)
	{
	case kControlUpButtonPart:
		if (value>0)
			SetControlValue(control,value-1);
		break;
	case kControlDownButtonPart:
		if (value<GetControlMaximum(control))
			SetControlValue(control,value+1);
		break;
	case kControlPageUpPart:
		SetControlValue(control,value-8);
		break;
	case kControlPageDownPart:
		SetControlValue(control,value+8);
		break;
	case kControlIndicatorPart:
		break;
	}

	if ((value-=GetControlValue(control))!=0 /* sub-ing new; negative = forward */
			&& (window=GetControlOwner(control))!=NULL)
	{ /* scroll window and update */
		Str255	title;

		GetControlTitle(control,title);
		if (title[1]=='V')
			listwindow_scrollbox(window,0,value*listwindow_getlheight(window));
		else
			listwindow_scrollbox(window,6*value,0);
	}

	if (!reference)
		SetControlReference(control,ticks+SCROLL_DOWN_DELAY);
}

static void alignbounds(int lheight,
	long newdimensions,
	Rect *bounding)
{
	*(int32*)bounding = 0;
	bounding->right = (int16)newdimensions;

	/* make height divisible by the line height */
	newdimensions+= 4+(lheight/2)<<16;
	newdimensions/= lheight<<16;
	newdimensions*= lheight;
	bounding->bottom = (newdimensions-=4);
}

WindowRef listwindow_create(WindowRef behind,
	const Rect *bounds,
	const uint8 *title,
	const uint8 *fontfamily,
	int16 textsize,
	int16 maxlsize,
	int32 maxlcount,
	int16 bottompad,
	int sel,
	long refCon)
{
	FontInfo	fontinfo;
	Rect		ctlbounds;
	WindowRef	window;
	GrafPtr		oldport;
	Ptr		storage;
	int		lh;
	int16		fontid, height, width;

	GetPort(&oldport);

	if ((storage=NewPtrClear(256))==NULL)
		return 0;
	if ((window=NewCWindow(storage,bounds,title,false,
			bottompad<0 ? movableDBoxProc:documentProc,
			behind,true,refCon))==NULL)
	{
		DisposePtr(storage);
		return 0;
	}

	setportwindow(window);
	if (fontfamily)
		GetFNum(fontfamily,&fontid);
	TextFont(fontid);
	TextSize(textsize);
	TextMode(srcCopy); /* white background */
	/* calculate line height */
	GetFontInfo(&fontinfo);
	if ((lh=fontinfo.ascent+fontinfo.descent)==0) //+1)==0)
		lh = 11; //12;
	LISTWINDOW_LHGT(window) = lh;
	if (sel>=0)
		LISTWINDOW_FLAG(window) = sel ? LW_SLECTABLE|LW_MULTIPLESEL:LW_SLECTABLE;
	if (bottompad<0)
		LISTWINDOW_FLAG(window)|= LW_MODAL;
	LISTWINDOW_SELEND(window) =-1;
	/* stuff in the rest of the parameters */
	LISTWINDOW_INDENT(window).h = DEFAULT_INDENT_H;
	LISTWINDOW_INDENT(window).v = DEFAULT_INDENT_V;
	LISTWINDOW_BOTTOMPAD(window) = (bottompad>>15)^bottompad;

	alignbounds(lh,(bounds->bottom-bounds->top<<16)|(uint16)(bounds->right-bounds->left),&ctlbounds);
	SizeWindow(window,width=ctlbounds.right,ctlbounds.bottom,false);
	height = ctlbounds.bottom-LISTWINDOW_BOTTOMPAD(window);
	LISTWINDOW_MAXLCOUNT(window) = ABS(maxlcount);
	if (maxlcount>=0)
	{
		ctlbounds.top =-1;
		ctlbounds.left = width-15;
		ctlbounds.bottom = height-14;
		ctlbounds.right = width+1;
		LISTWINDOW_VSCROLL(window) = NewControl(window,&ctlbounds,"\pV",true,
			0,0,0,scrollBarProc,0);
	}
	if ((LISTWINDOW_MAXLSIZE(window)=maxlsize)!=0)
	{
		ctlbounds.top = height-15;
		ctlbounds.left =-1;
		ctlbounds.bottom = height+1;
		ctlbounds.right = width-14;
		LISTWINDOW_HSCROLL(window) = NewControl(window,&ctlbounds,"\pH",true,
			0,0,MAX(maxlsize,64)-64,scrollBarProc,0);
	}

	MacSetPort(oldport);

	return window;
}

int listwindow_dispose(WindowRef window)
{
	if (!window)
		return -1;
	listwindow_free(window);
	KillControls(window);
	DisposeWindow(window);
	return 1;
}

int listwindow_update(WindowRef window,
	Boolean erase)
{
	Rect		bounds;
	GrafPtr		oldport;
	int		r;

	if (!getvisiblewindowport(window))
		return -1;
	getwindowbounds(window,&bounds);

	GetPort(&oldport);
	setportwindow(window);

	ClipRect(&bounds);
	DrawControls(window);

	if (LISTWINDOW_VSCROLL(window))
		bounds.bottom-= 15;
	if (LISTWINDOW_HSCROLL(window))
		bounds.right-= 15;
	ClipRect(&bounds);

	if (erase)
		EraseRect(&bounds);

	--bounds.bottom, --bounds.right; /* not prep'd for full inset (drawing errors) */
	ClipRect(&bounds);

	r = listwindow_drawtext(window,(bounds.bottom-bounds.top)/listwindow_getlheight(window));

	bounds.bottom+= 16, bounds.right+= 16;
	ClipRect(&bounds);

	if (erase==true && !(LISTWINDOW_FLAG(window)&LW_MODAL))
		drawgrowicon(window);

	MacSetPort(oldport);

	return r;
}

int listwindow_grow(WindowRef window,
	const EventRecord *event,
	Rect *bounding)
{
	long		newdimensions;

	if (!getvisiblewindowport(window))
		return -1;
	/* GrowWindow subtracts one when min=max, can adjust wholey via command key */
	if ((newdimensions=GrowWindow(window,event->where,bounding))!=0)
	{
		alignbounds(listwindow_getlheight(window),newdimensions,bounding);
		SizeWindow(window,bounding->right,bounding->bottom,false);
		listwindow_updatescrolls(window,true); /* this was missing for so long!; 2011.03.24 */
		/* because of bugs in SizeWindow, invalidation should be done manually */
		listwindow_update(window,true);

		return 1;
	}
	return 0;
}

//#include <stdio.h>

#define FUDGE_V				3

static int listwindow_clickselect(WindowRef window,
	EventRecord *event,
	Point hitpoint)
{
	Rect		bounds;
	GrafPtr		oldport;
	int		x, y, t, l, m, lheight = listwindow_getlheight(window);
	int		lt, nt;
	Boolean		ms;

	getlistbounds(window,&bounds);
	listwindow_getscrolls(window,&x,&y);
	if ((t=LISTWINDOW_INDENT(window).h-6*x)<0)
		t = 0;
	bounds.left = t;
	t = LISTWINDOW_INDENT(window).v-lheight*y;
	bounds.top = MAX(t,0);
	if (hitpoint.h<bounds.left || hitpoint.v<bounds.top
			|| hitpoint.h>=bounds.right || hitpoint.v>=bounds.bottom)
		return 0;
	//printf("t=%d,v=%d,h=%d\n",
	//	t,hitpoint.v,lheight);
	if ((y=(hitpoint.v-t-FUDGE_V)/lheight)<0 || y>=LISTWINDOW_LCOUNT(window))
		return 0;
	ms = LISTWINDOW_FLAG(window)&LW_MULTIPLESEL && event->modifiers&shiftKey;
	/* direction from middle */
	l =-1, m = LISTWINDOW_SELSTART(window)+LISTWINDOW_SELEND(window)>>1;
	m = y>m ? LISTWINDOW_SELSTART(window):LISTWINDOW_SELEND(window);
	lt = TickCount();

	GetPort(&oldport);
	setportwindow(window);

	do
	{
		if (l!=y)
		{
			l = y;
			if (!ms)
				m = y;
			if (listwindow_select(window,MIN(y,m),MAX(y,m)))
				listwindow_drawtext(window,(bounds.bottom-bounds.top)/lheight);
		}
		GetMouse(&hitpoint);
		if ((hitpoint.v<bounds.top || hitpoint.v>bounds.bottom) && (nt=TickCount())-lt>=0)
		{
			if ((t=hitpoint.v-bounds.top)<0)
				t =-t, y =-1;
			else
				t = hitpoint.v-bounds.bottom, y = 1;
			lt = nt+32/PIN(t,2,32);
			listwindow_scroll(window,0,y);
			listwindow_getscrolls(window,&x,&y);
			t = LISTWINDOW_INDENT(window).v-lheight*y;
			goto upy;
		}
		if (hitpoint.h>=bounds.left && hitpoint.h<bounds.right)
		upy:	y = (hitpoint.v-t-FUDGE_V)/lheight;
	} while(StillDown());

	SetPort(oldport);

	return listwindow_getsel(window,NULL,NULL);
}

int listwindow_click(WindowRef window,
	EventRecord *event,
	short clickpart,
	Rect *growbounding)
{
	Point		hitpoint = event->where;
	ControlRef	control;

	ControlPartCode		part;

	if (!getvisiblewindowport(window))
		return -1;
	localwindowpoint(window,&hitpoint);

	if (clickpart==inGrow && growbounding)
		return listwindow_grow(window,event,growbounding);
	/* both scrolls show reflect the same 'visible' */
	if ((control=LISTWINDOW_VSCROLL(window))!=NULL && !IsControlVisible(control))
		listwindow_setactive(window,true); /* manual update; 2010.05.27 */

	if (!lwctlupp && (lwctlupp=NewControlActionUPP(mycontrolactionproc))==NULL)
		return -2;
	if ((part=FindControl(hitpoint,window,&control))==kControlNoPart || !control)
	{
		if (LISTWINDOW_FLAG(window)&LW_SLECTABLE)
			return listwindow_clickselect(window,event,hitpoint);
		return 0;
	}
	SetControlReference(control,0); /* initialize timing for control */

	switch(part)
	{
	case kControlUpButtonPart:
	case kControlDownButtonPart:
	case kControlPageUpPart:
	case kControlPageDownPart:
		TrackControl(control,hitpoint,lwctlupp);
		break;
	case kControlIndicatorPart:
		TrackControl(control,hitpoint,NULL);
		listwindow_update(window,2);
	}
	return 1;
}

int listwindow_setactive(WindowRef window,
	Boolean active)
{
	ControlRef	control;

	if (!getvisiblewindowport(window))
		return 0;
	if ((control=LISTWINDOW_VSCROLL(window))!=NULL)
		ShowHideControl(control,active);
	if ((control=LISTWINDOW_HSCROLL(window))!=NULL)
		ShowHideControl(control,active);
	if (!(LISTWINDOW_FLAG(window)&LW_MODAL))
		drawgrowicon(window);

	HiliteWindow(window,active);

	return 1;
}

static int32 gettoffset(WindowRef window,
	int32 x)
{
	const int32	*a;
	int32		i;

	if ((a=LISTWINDOW_TLINELIST(window))==NULL || (i=LISTWINDOW_TFIRSTLN(window))<0)
		return -1;
	if (LISTWINDOW_MAXLCOUNT(window))
		a+= i;
	return x<0 ? 0:x>=LISTWINDOW_LCOUNT(window) ? LISTWINDOW_TLENGTH(window):a[x];
}

int listwindow_copy(WindowRef window)
{
	int32		s, e;

	if (!window)
		return -1;
	if (!LISTWINDOW_TEXT(window)
			|| (s=gettoffset(window,LISTWINDOW_SELSTART(window)))<0
			|| (e=gettoffset(window,LISTWINDOW_SELEND(window)+1))<0
			|| s>=e)
		return 0;
#if CALL_NOT_IN_CARBON
	ZeroScrap();
	PutScrap(e-s,kScrapFlavorTypeText,s+LISTWINDOW_TEXT(window));
#else
	ClearCurrentScrap();
	{
		ScrapRef	scrap;

		GetCurrentScrap(ScrapRef *scrap);
		PutScrapFlavor(scrap,kScrapFlavorTypeText,kScrapFlavorMaskNone,
			e-s,s+LISTWINDOW_TEXT(window));
	}
#endif
	return 1;
}
