/*
 macos/macsound.c
  sound for MacFCEU

  Created: 2008.07.02
  Revised: 2013.05.14

 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Tuesday, September 16, 2008 10:52 PM EDST
  because playback end starts at buffered end, double buffering can be removed
Tuesday, January 19, 2009 11:43 PM
  sound rate NTSC mod (ConsoleDownRate) moved to sound.c.
Saturday, October 24, 2009 9:08 PM
  officially getting rid of the non-68k callback stuff.
*/

#include <stdio.h>

#include <Sound.h>

#include "shell.h"

#include "../FCEU.h"
#include "../utility/ftrig.h"
#include "../common/dface.h"
#include "../WAVE.h"  /* for WAVE_DefaultRate */

#include "shellmenus.h"

#ifdef env68k
 #pragma segment Sound
#endif

//#ifndef USE_DIRECT_COMPONENT_SOUND -- deprecated
/* to keep audio playing with gaps - slower performance */
#define STALL_SILENCE			0

/* ---- macros */

#define _bcpy2(d,s,c)			do{void *__d=d; const void *__s=s; long __z=c; \
	if(__z&1) *((uint16*)__d)++=*((uint16*)__s)++; \
	if(__z&2) *((uint32*)__d)++=*((uint32*)__s)++; \
	if(__z&4) *((uint32*)__d)++=*((uint32*)__s)++, \
		*((uint32*)__d)++=*((uint32*)__s)++; \
	if(__z&8) *((uint32*)__d)++=*((uint32*)__s)++, \
		*((uint32*)__d)++=*((uint32*)__s)++, \
		*((uint32*)__d)++=*((uint32*)__s)++, \
		*((uint32*)__d)++=*((uint32*)__s)++; \
	if ((__z&=-16)!=0) do{ \
		*((uint32*)__d)++=*((uint32*)__s)++,*((uint32*)__d)++=*((uint32*)__s)++; \
		*((uint32*)__d)++=*((uint32*)__s)++,*((uint32*)__d)++=*((uint32*)__s)++; \
		*((uint32*)__d)++=*((uint32*)__s)++,*((uint32*)__d)++=*((uint32*)__s)++; \
		*((uint32*)__d)++=*((uint32*)__s)++,*((uint32*)__d)++=*((uint32*)__s)++; \
	} while((__z+=-16)!=0);}while(0)

#if TARGET_CPU_68K
 #define __FIXEDEXT80(x)		((x)/65536.0)
#else
 #define __FIXEDEXT80(x)		{0x400E, x, 0}
#endif

/* ---- globals */

extern uint8		VMMask;

/* surpassing buffer size in sound.c (0x2EE0) */
#define MAXIMUM_RATE			96000
#define BUFFER_SIZE			0x3000 /* must be >minbufsize */
#define BUFFER_COUNT			4 /* must be a power of two */
#define BUFFER_MASK			(BUFFER_COUNT-1)
#define BUFFER_MAXIMUM			(BUFFER_SIZE-32)

static int16	playbuffer[BUFFER_COUNT][BUFFER_SIZE];

static SndCallBackUPP	userroutine;

SndChannel	mychannel =
{
	NULL, NULL, NULL, /* next channel, first mod, callback proc */
	0, 0, /* user info (buffer index and mask),  wait */
	{0}, /* command in progress */
	0, stdQLength, 0, 0, /* flags, qLength, qHead, qTail */
};
#define bufindexread			*((volatile uint16*)&mychannel.userInfo)
#define bufindexwrite			((uint16*)&mychannel.userInfo)[1]

static CmpSoundHeader	myheader =
{
	0, /* sample data */
	1, rate22khz, 0, 0, /* channel count, sample rate, loop start, end */
	cmpSH, 60, /* encoding, base note */
	0, __FIXEDEXT80(rate22khz), /* frame count, floating sample rate */
	NULL, /* marker sync track */
	k16BitNativeEndianFormat,
	0, NULL, NULL, 0, /* reserved, state block ptr, leftover samples, codec ID */
	0, 0, 16, /* packet size, synth ID, sample size */
};
static int		mylengths[BUFFER_COUNT];//, minwaitcount;

/* ---- code */

int KillSound()
{
	SndCommand	command;

	command.cmd = bufferCmd;
	command.param1 = 0;
	command.param2 = (long int)&myheader;

	if (!mychannel.firstMod || !myheader.samplePtr)
		return 0;
	memset(playbuffer,0,16);
	mychannel.userInfo = 0;
	myheader.numFrames = 16;
	SndDoImmediate(&mychannel,&command);

	SndDisposeChannel(&mychannel,true);
	mychannel.firstMod = 0;
#if CALL_NOT_IN_CARBON
	if (VMMask&1)
		UnlockMemory(playbuffer,sizeof(playbuffer));
#endif
	DisposeSndCallBackUPP(userroutine); userroutine = NULL;

	return 1;
}
/* -- doesn't work!
static void MyInsertCommand(SndChannel *chan,
	int16 cmd,
	void *ptr)
{
	SndCommand	*qcmd;

	if (chan->qHead==-1)
		chan->qHead = chan->qTail;
	if (chan->qHead>=1)
		--chan->qHead;
	else
		chan->qHead = chan->qTail;
	qcmd = chan->queue+chan->qHead;
	qcmd->cmd = cmd;
	qcmd->param1 = 0;
	qcmd->param2 = (long int)ptr;
}
*/
static pascal void mychannelcb(SndChannelPtr chan,
	SndCommand *cmd)
{
	int		ix = bufindexread&BUFFER_MASK;
#pragma unused(cmd)
	if (!chan->userInfo)
		return;
	myheader.samplePtr = (char*)playbuffer[ix];
	if (ix==(bufindexwrite&BUFFER_MASK)
			|| (myheader.numFrames=mylengths[ix])==0)
		return;
	else
	{
		myheader.numFrames = mylengths[ix];
		mylengths[ix] = 0; /* no replay */
		++ix, bufindexread = ix&BUFFER_MASK;
	}
	cmd->cmd = bufferCmd;
	cmd->param2 = (long int)&myheader;
	SndDoCommand(&mychannel,cmd,0);
	cmd->cmd = callBackCmd;
	cmd->param2 = 0;
	SndDoCommand(&mychannel,cmd,0);
}

/* null git now taken as normal */
/* (soundon/soundvolume determine if channel should be created) */
int InitSound(FCEUGI *git)
{
	SndChannel	*mychannelptr = &mychannel;
	UnsignedFixed	samplerate, ratemultiplier = 0x10000u;
	int		chancnt;
	OSErr		err;

	if (mychannel.firstMod)
		KillSound();

	if (!soundon || !soundvolume) /*git==NULL)*/
		{FCEUI_Sound(0); return 0;}

	/* fix sample rate variables */
	if (!git)
		samplerate = sndset.rate;
	else if ((samplerate=git->soundrate)==0)
		git->soundrate = samplerate = sndset.rate;
	else if (sndset.rate==0)
		sndset.rate = samplerate;

	if (samplerate<1000 || samplerate>MAXIMUM_RATE)
	{ /* fix range */
		samplerate = sndset.rate = samplerate<1000 ? 1000:MAXIMUM_RATE;
		updatemenus();
		FCEUI_SetSoundQuality(0);
	}

	chancnt = git && git->soundchan>=2 ? 2:1;
//#if STALL_SILENCE
//	stallsize = PIN((unsigned)sndset.rate,2048*8,2048*32)>>11;
//#endif
	FCEUI_Sound(WAVE_DefaultRate=samplerate);
	FCEUI_SetSoundActiveChannels(sndactivechannels);

	if (!soundavailable)
		return 1;
#if CALL_NOT_IN_CARBON
	if (VMMask&1)
		LockMemory(playbuffer,sizeof(playbuffer));
#endif
	err =-1;
	if ((userroutine=NewSndCallBackUPP(mychannelcb))==NULL
			|| (err=SndNewChannel(&mychannelptr,sampledSynth,chancnt==2 ?
			initStereo:initMono,userroutine))!=noErr)
	{
		if (err==-32767 || err==noHardwareErr)
		{
			soundavailable = 0;
			mychannel.firstMod = 0; /* not automatically cleared */
			FCEU_printf("## InitSound: no sound device for sound channel\n");
		} else
			FCEU_printf("## InitSound: failed to create sound channel (#%d)\n",
				err);
		return -1;
	}

	{ /* support for sample rates >= 65536 */
		int		shift;

		if ((shift=cntlz32(samplerate))>15)
			samplerate<<= 16;
		else if ((ratemultiplier<<=16+-shift)==0)
			{ratemultiplier =-1ul; /*DebugStr("\pDemon, be Gone!");*/}
		else
			samplerate<<= shift;
	}
	myheader.samplePtr = (char*)playbuffer;
	myheader.numChannels = chancnt;
	myheader.sampleRate = samplerate;
	if (ratemultiplier!=0x10000ul)
	{
		SndCommand	command;

		command.cmd = rateMultiplierCmd;
		command.param1 = 0;
		command.param2 = ratemultiplier;
		err = SndDoImmediate(&mychannel,&command);
	}
	/* throttle always on now; 2012.01.03 04:46 ET */
	RefreshThrottleFPS();

	//minwaitcount = samplerate/61;

	return 1;
}

void writesound(int16 *data,
	int count)
{
extern int32 frametime;
	SndCommand	command;
	int		ix;

	if (!mychannel.firstMod || count<=0)
		return;
	//if (NoWaiting&1) /* sound skipping now handled by shell.c */
	//	bufindexread = bufindexwrite; /* reset */
	if (!data)
		return;
	if (count>BUFFER_MAXIMUM)
		count = BUFFER_MAXIMUM;
	/* faster than a 61Hz frame won't play full speed (emulated) anyway */
	if (frametime>=(32000000/61) && bufindexwrite)
	{ /* wait till enough is finished - max. 1/30 to 4.25 sec. */
		int	ticks = 2+TickCount(); /* 4 didn't work */

		while(((uint16)(bufindexwrite-bufindexread)&BUFFER_MASK)>2
		//while(prevQ==mychannel.qHead && mychannel.qHead!=-1
				&& (int8)(TickCount()-ticks)<0) /* overflow protected */
			;
	}
	command.param1 = 0;
	mylengths[ix=bufindexwrite&BUFFER_MASK] = count;

	/* copy the whole thing, play it in chunks */
	_bcpy2((char*)playbuffer[ix],data,count);
	/* SndDoImmediate with this multi-buffering form hangs */
	command.cmd = callBackCmd;
	command.param2 = 0;
	/* increment and active */
	++ix, bufindexwrite = ix|0x8000u;
	if (mychannel.qHead==-1)
		mychannelcb(&mychannel,&command);
}

void stopsound()
{
	SndCommand	command;

	if (!mychannel.firstMod)
		return;
	command.cmd = quietCmd;
	command.param1 = 0;
	command.param2 = 0;
	SndDoImmediate(&mychannel,&command);
	command.cmd = flushCmd;
	SndDoImmediate(&mychannel,&command);

	mychannel.userInfo = 0;
}

boolean adequatesoundmanager()
{
	NumVersion	version = SndSoundManagerVersion();

	return version.majorRev>=3;
}

int adjust_soundvolume(int displacement)
{
	int		volume; /* shouldnt directly set the global */

	if (!soundon)
		return -1;
	if ((volume=soundvolume/10)<0)
		volume = 16;
	else if ((volume+=displacement)<0)
		volume = 0;
	else if (volume>15)
		volume = 16;
	return domenu(_subvolume_off+volume); /* should call InitSound during 0% volume game */
}
