/*
 palette.c

 * MacFCEU 0.6.4
 *
 * Copyright notice for the original file:
 *  palette.c Copyright (C) 2002,2003 Xodnizel
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>
#include <stdlib.h>

#include "driver.h"
#include "FCEU.h"
#include "general.h"
#include "palette.h"
#include "palettes/palettes.h" /* keep this here as to not intefere w/ <Palettes.h> */

#include "utility/ftrig.h"

uint8		pale = 0;
static uint8	pali = 0;
static uint8	prevemph = 0;

static uint8	ntsccol = 0;
static int	ntsctint = 46+10; //i'm using 72, 56...
static int	ntschue = 72;
static int	ntscbrt = 64; /* added; 2010.05.09 */

/* These are dynamically filled/generated palettes: */
pal	palettei[64]; // Loaded palette for an individual game (based on its name).
pal	palettec[64]; // Custom "global" palette.
pal	paletten[64]; // Mathematically generated NTSC palette.

pal		*palo;

static pal *palpoint[8] =
{
	palette,
	rp2c04001,
	rp2c04002,
	rp2c04003,
	rp2c05004,
};

void FCEUI_SetPaletteArray(uint8 *p)
{
	if (p==0)
		palpoint[0] = palette;
	else
	{
		pal	*ent = palettec;
		int	i = 64;

		palpoint[0] = ent;
		do
			ent->r = *p++,
			ent->g = *p++,
			ent->b = *p++;
		while(++ent,--i!=0);
	}
	FCEU_ResetPalette(-1);
}


void FCEUI_SetNTSCTHB(int n,
	int tint,
	int hue,
	int brt)
{
	ntsccol = n!=0;
	ntsctint = tint;
	ntschue = hue;
	ntscbrt = brt;
	FCEU_ResetPalette(1); //always recalculate here; 2009.01.29
}

void SetNESDeemph(uint8 d,
	int force)
{ //only 2C07/2C02 treat this as a 1.0:0.75|0.85 form; the other 2C0x's make a white-wash out
	static uint16	rtmul[7] =
		{32768*1.239,32768*0.794,32768*1.019,32768*0.905,32768*1.023,32768*0.741,32768*.75};
	static uint16	gtmul[7] =
		{32768*0.915,32768*1.086,32768*0.980,32768*1.026,32768*0.908,32768*0.987,32768*.75};
	static uint16	btmul[7] =
		{32768*0.743,32768*0.882,32768*0.653,32768*1.277,32768*0.979,32768*1.001,32768*.75};

	const pal	*s;
	uint32		r, g, b;
	int		x;

	if (palo==0)
		return;

	/* If it's not forced (only forced when the palette changes),
		don't waste cpu time if the same deemphasis bits are set as the last call.
	*/
	if (!force)
	{
		if (d==prevemph)
			return;
	} else
	{ /* Only set this when palette has changed. */
		r = rtmul[6];
		g = rtmul[6];
		b = rtmul[6];
		x = 0xC0, s = palo;
		do
		{
			uint32		m = s->r, n = s->g, o = s->b;

			if ((m=(m*r)>>15)>255)
				m = 255;
			if ((n=(n*g)>>15)>255)
				n = 255;
			if ((o=(o*b)>>15)>255)
				o = 255;
			FCEUD_SetPalette(x,m,n,o);
		} while(++s,++x<0x100);
	}

	if (d==0)
		return; //No emphasis

	r = rtmul[d-1];
	g = gtmul[d-1];
	b = btmul[d-1];
	x = 0x40, s = palo; //colors dependent on 0x40-0x7F will screw up
	do
	{
		uint32		m = s->r, n = s->g, o = s->b;

		if ((m=(m*r)>>15)>255)
			m = 255;
		if ((n=(n*g)>>15)>255)
			n = 255;
		if ((o=(o*b)>>15)>255)
			o = 255;
		FCEUD_SetPalette(x,m,n,o);
	} while(++s,++x<0x80);

	prevemph = d;
}

static void WritePalette(void)
{
	int x;

	for(x=0;x<7;x++)
		FCEUD_SetPalette(x,unvpalette[x].r,unvpalette[x].g,unvpalette[x].b);
	if (FCEUGameInfo->type==GIT_NSF)
	{
	//for(x=0;x<128;x++)
	// FCEUD_SetPalette(x,x,0,x);
	} else
	{
		for(x=0; x<64; x++)
			FCEUD_SetPalette(128+x,palo[x].r,palo[x].g,palo[x].b);
		SetNESDeemph(prevemph,1);
	}
}

//CalculatePalette here is being revised.
//...What's written in Nintendulator/GFX.cpp isn't accurate after all.
// nes_ntsc (by Shay Green) is the one to look up to.

#define W		255		//chromatic white/full (8bpc)

#define YS		7		//luminescence sub-fractional bits
#define YW		(W<<YS)		//scaled white
#define DS		9		//decode fractional bits
#define DM		((1<<DS)*5/4)

#define TM		(1<<32-YS-DS)	//output scaler

uint32		CalcPOutscales[3] = {1.0*TM,1.0*TM,1.0*TM}; /*can be adjusted externally*/

int16		CalcPDecodeIQ[3][2] =
{ /* these are supposed to be correct, but greens are bright and reds are dim */
	{ 0.956*DM, 0.621*DM},
	{-0.272*DM,-0.647*DM},
	{-1.105*DM, 1.702*DM}
};

/* the entries can only barely exceed +/-1.0 */
/* some too dark, some too bright */
int16		CalcPHiY[4] = {0.40*YW,0.68*YW,1.00*YW,1.00*YW};
int16		CalcPLoY[4] = {-.12*YW,0.00*YW,0.31*YW,0.72*YW};
/* the third here makes for the difference of the above two */
int16		CalcPDfY[4] = {0.52*YW,0.68*YW,0.69*YW,0.28*YW};

static void CalculatePalette(void)
{

	pal		*ent;
	unsigned	S, sat;
	ufract		H, hue;
	long		brt = ntscbrt-64<<YS+DS+1;
	int		x, z, v;

	if ((sat=ntsctint)<0)
		sat = 0;
	else if (sat>9999)
		sat = 10000;
	hue = (long)(4294967296.0/5)+umuldiv(ntschue%360+360,4294967296.0/2,360);

	for(x=0,ent=paletten; x<4; ++x)
		for(z=0,H=hue; z<16; ++z,++ent)
		{
			long	Y, I = 0, Q = 0;

			if (z==0)
				Y = CalcPHiY[x]<<DS;
			else if (z<13)
				Y = (CalcPLoY[x]<<DS)+(CalcPDfY[x]<<DS-1),
				S = (CalcPDfY[x]*sat)>>7,
				I = ifract30mul(S,fastsin(H)),
				Q = ifract30mul(S,fastcos(H)),
				H+= (ufract)(30.0*(double)4294967296.0/360);
			else if (z==13)
				Y = CalcPLoY[x]<<DS;
			else
				Y = 0;
			Y+= brt;
			if ((v=Y+I*CalcPDecodeIQ[0][0]+Q*CalcPDecodeIQ[0][1]+(1<<YS-1))<0)
				v = 0;
			else if ((v=ufract32mul(v,CalcPOutscales[0]))>W)
				v = W;
			ent->r = v;
			if ((v=Y+I*CalcPDecodeIQ[1][0]+Q*CalcPDecodeIQ[1][1]+(1<<YS-1))<0)
				v = 0;
			else if ((v=ufract32mul(v,CalcPOutscales[1]))>W)
				v = W;
			ent->g = v;
			if ((v=Y+I*CalcPDecodeIQ[2][0]+Q*CalcPDecodeIQ[2][1]+(1<<YS-1))<0)
				v = 0;
			else if ((v=ufract32mul(v,CalcPOutscales[2]))>W)
				v = W;
			ent->b = v;
		}
	WritePalette();
}

static int ChoosePalette(void)
{
	if (FCEUGameInfo->type==GIT_NSF)
		palo = 0;
	else if (pali)
		palo = palettei;
	else if (ntsccol && !PAL && FCEUGameInfo->type!=GIT_VSUNI)
	{
		palo = paletten;
		CalculatePalette();
		return 0; /* WritePalette already called */
	} else
		palo = palpoint[pale];

	return 1;
}

void FCEU_LoadGamePalette(void)
{
	uint8	ptmp[192];
	FILE	*fp;
	char	*fn;

	pali = 0;

	if ((fn=FCEU_MakeFName(FCEUMKF_PALETTE,0,0))==0)
		return;

	if ((fp=FCEUD_UTF8fopen(fn,"rb"))!=0)
	{
		const uint8	*p;
		int		x, z;

		if ((z=fread(ptmp,3,64,fp))>64)
			z = 64; //use the amount read; clip to 64 color entries
		fclose(fp);

		for(x=0,p=ptmp; x<z; ++x)
			palettei[x].r = *p++,
			palettei[x].g = *p++,
			palettei[x].b = *p++;
		for(; x<64; ++x) //clear out the rest if incomplete
			palettei[x].r =
			palettei[x].g =
			palettei[x].b = 0;

		pali = 1;
	}
	free(fn);
}

void FCEU_ResetPalette(int set)
{
	if ((FCEUGameInfo || set>0) && set!=0)
		if (ChoosePalette())
			WritePalette();
}

void FCEUI_GetNTSCTHB(int *tint,
	int *hue,
	int *brt)
{
	*tint = ntsctint;
	*hue = ntschue;
	*brt = ntscbrt;
}

static int	*ctrlselect = 0;
static uint16	ctrltime = 0;

void FCEUI_NTSCDEC(void)
{
	if (ntsccol && !PAL && FCEUGameInfo->type!=GIT_VSUNI
			&& FCEUGameInfo->type!=GIT_NSF && ctrlselect)
	{
		if (ctrltime)
		{
			if (--(*ctrlselect)<0)
				*ctrlselect = 0;
			CalculatePalette();
		}
		ctrltime = 360;
	}
}

void FCEUI_NTSCINC(void)
{
	if (ntsccol && !PAL && FCEUGameInfo->type!=GIT_VSUNI
			&& FCEUGameInfo->type!=GIT_NSF && ctrlselect)
	{
		if (ctrltime)
		{
			if (++(*ctrlselect)>128)
				*ctrlselect = 128;
			CalculatePalette();
		}
		ctrltime = 360;
	}
}

void FCEUI_NTSCSelect(uint8 w)
{
	if (ntsccol && !PAL && FCEUGameInfo->type!=GIT_VSUNI
			&& FCEUGameInfo->type!=GIT_NSF)
	{
		switch(w)
		{
		case 0: ctrlselect = &ntsctint; break;
		case 1: ctrlselect = &ntschue; break;
		case 2: ctrlselect = &ntscbrt; break;

		default: return;
		}
		ctrltime = 360;
	}
}

void FCEU_DrawNTSCControlBars(uint8 *XBuf)
{
	int		x, x2, which = 0;

	if (XBuf==0)
		ctrltime = 0;
	if (ctrltime==0)
		{ctrlselect = 0; return;}
	--ctrltime;

	if (ctrlselect==&ntschue)
		DrawTextTrans(XBuf+128-12+180*256,256,(uint8*)"Hue",0x85);
	else if (ctrlselect==&ntsctint)
		DrawTextTrans(XBuf+128-16+180*256,256,(uint8*)"Tint",0x85);
	else if (ctrlselect==&ntscbrt)
		DrawTextTrans(XBuf+128-40+180*256,256,(uint8*)"Brightness",0x85);
	which = *ctrlselect*2;

	XBuf+= 200*256;
	for(x=0; x<which; x+=2)
		for(x2=6; x2>=-6; --x2)
			XBuf[x-256*x2] = 0x85;
	for(; x<256; x+=2)
		for(x2=2; x2>=-2; --x2)
			XBuf[x-256*x2] = 0x85;
}
