/*
 utility/disasm6502.h

  Revised: 2013.05.09

 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

enum
{ /* index displacement alters only the low byte */
	OPM_IM, /* 1-immediate		- #$xx */
	OPM_RL, /* 1-relative		- *+$xx */
	OPM_ZP, /* 1-zero page		- $xx */
	OPM_ZX, /* 1-zero page, X index	- $xx,X */
	OPM_ZY, /* 1-zero page, Y index	- $xx,Y */
	OPM_IX, /* 1-indirect, X index	- ($xx,X) */
	OPM_IY, /* 1-indirect, Y index	- ($xx),Y */
	OPM_IN, /* 1-indirect		- ($xx) */
	OPM_IB, /* 2-absolute indirect	- ($xxxx) */
	OPM_AB, /* 2-absolute		- $xxxx */
	OPM_AX, /* 2-absolute, X index	- $xxxx,X */
	OPM_AY, /* 2-absolute, Y index	- $xxxx,Y */
	OPM_IA, /* 0-implied A to self */
	OPM_IP, /* 0-implied */

	OPM_COUNT /* 14 */
};

enum
{ /* all opcodes are defined */
	OPN_NOP, /* 0 - official NOP is defined here */
	OPN_ZOP, /* "act but don't affect any visible registers" */
	OPN_KIL,
	OPN_BRK,
	OPN_JMP, /* 4 */
	OPN_JSR,
	OPN_RTI,
	OPN_RTS,
	OPN_PHA, /* 8 */
	OPN_PHP,
	OPN_PLA,
	OPN_PLP,
	OPN_SHA, /* 12 */
	OPN_SHY,
	OPN_SHX,
	OPN_SHS,
	OPN_BPL, /* 16 */
	OPN_BMI,
	OPN_BVC,
	OPN_BVS,
	OPN_BCC, /* 20 */
	OPN_BCS,
	OPN_BNE,
	OPN_BEQ,
	OPN_STX, /* 24 */
	OPN_STY,
	OPN_STA,
	OPN_SAX,
	OPN_TXS, /* 28 */
	OPN_TSX, /* 29 - first opc that affects P register */
	OPN_ANE,
	OPN_ANC,
	OPN_CLC, /* 32 */
	OPN_CLI,
	OPN_CLV,
	OPN_CLD,
	OPN_SEC, /* 36 */
	OPN_SEI,
	OPN_SED,
	OPN_BIT,
	OPN_TXA, /* 40 */
	OPN_TAX,
	OPN_TYA,
	OPN_TAY,
	OPN_DEX, /* 44 */
	OPN_DEY,
	OPN_INX,
	OPN_INY,
	OPN_DEC, /* 48 */
	OPN_INC,
	OPN_SLO,
	OPN_SRE,
	OPN_CPX, /* 52 */
	OPN_CPY,
	OPN_CMP,
	OPN_DCP,
	OPN_ADC, /* 56 */
	OPN_SBC,
	OPN_SBX,
	OPN_ISB,
	OPN_LAX, /* 60 */
	OPN_LAS,
	OPN_LDX,
	OPN_LDY,
	OPN_LDA, /* 64 */
	OPN_ORA,
	OPN_AND,
	OPN_EOR,
	OPN_ASL, /* 68 */
	OPN_LSR,
	OPN_ARR,
	OPN_ASR,
	OPN_ROL, /* 72 */
	OPN_ROR,
	OPN_RLA,
	OPN_RRA,

	OPN_COUNT /* 76 - everything's defined */
};

/* x6502_opn_table entry: [read.1] [write.1] [flag.1] [bad.1] [opm.4] [opn.8] */
#define OPB_READ		0x8000
#define OPB_WRITE		0x4000
#define OPB_FLAG		0x2000 /* instruction affects P reg. */
#define OPB_BAD			0x1000

extern const char	x6502_opsizes[16];
extern const char	x6502_opnametable[OPN_COUNT][3]; /* 228 bytes */
extern const uint16	x6502_opn_table[256]; /* 512 bytes */

/* NOTE: unlike FCEUI_Disassemble, fceuindbg is not affected (set yourself) */
uint32 X6502_Disassemble(const struct X6502struct *X, /* can be NULL (no peeking) */
	const readfunc *readers, /* must be defined (pass ARead) */
	uint32 addr, /* starting address to disassemble; next one returned */
	uint32 amask, /* for 16-bit addressing, pass 0xFFFF */
	int rdownshift, /* pass zero for one addr for each "reader" */
	char *strout); /* should be able to buffer >=34 bytes */
