/*
 ftrig.h
  for fractional integer routines (alternative to floating-point)

  Created: 2008.07.15
  Revised: 2013.05.15 - added fractasin

13.05.10 - fixed POCC __inline _cntlz32
13.03.30
13.01.08 - ifix16div (handles overflow; no rounding)
11.06.24 - moved types to enviro64.h, added ufix16mul
*/

#define ifix16div(x,y) (ifixed)((ifixed)(x)*65536.0/(ifixed)(y))

#ifndef __FTRIG_H
 #ifdef env68k
  #pragma parameter __D0 ufract32mul(__D0, __D1)
  extern uint32 ufract32mul(uint32, uint32) = {
/*	MULU.L	D0,D0:D1 */	0x4C00, 0x1400
  };
  /* overflow will yield zero, as 0x100000000 would yield the same */
  #pragma parameter __D0 ufract32div(__D1,__D2)
  extern uint32 ufract32div(uint32, uint32) = {
/*	MOVEQ	#0,D0	*/	0x7000,
/*	DIVU.L	D2,D1:D0 */	0x4C42, 0x401
  };

  #pragma parameter __D0 ifract30mul(__D0, __D1)
  extern int32 ifract30mul(int32, int32) = {
/*	MULS.L	D0,D0:D1 */	0x4C00, 0x1C00,
/*	ADD.L	D1,D1	*/	0xD281,
/*	ADDX.L	D0,D0	*/	0xD180,
/*	ADD.L	D1,D1	*/	0xD281,
/*	ADDX.L	D0,D0	*/	0xD180
  };
  #pragma parameter __D0 ufract30mul(__D0, __D1)
  extern uint32 ufract30mul(uint32, uint32) = {
/*	MULU.L	D0,D0:D1 */	0x4C00, 0x1400,
/*	ADD.L	D1,D1	*/	0xD281,
/*	ADDX.L	D0,D0	*/	0xD180,
/*	ADD.L	D1,D1	*/	0xD281,
/*	ADDX.L	D0,D0	*/	0xD180
  };

  #pragma parameter __D1 ifract30div(__D0,__D2)
  extern int32 ifract30div(int32, int32) = {
/*	MOVEQ	#3,D1	*/	0x7203,
/*	AND.L	D0,D1	*/	0xC280,
/*	ROR.L	#2,D1	*/	0xE499,
/*	ASR.L	#2,D0	*/	0xE480,
/*	MOVE.L	D2,-(A7) */	0x2F02,
/*	BPL.S	*+8	;@1 */	0x6A06,
/*	NEG.L	(A7)	*/	0x4497,
/*	NEG.L	D1	*/	0x4481,
/*	NEGX.L	D0	*/	0x4080,
/*@1	TDIVS.L	(A7),D2:D0 */	0x4C57, 0x802,
/*	DIVS.L	(A7)+,D2:D1 */	0x4C5F, 0x1C02
  };
  #pragma parameter __D1 ufract30div(__D0,__D2)
  extern uint32 ufract30div(uint32, uint32) = {
/*	MOVEQ	#3,D1	*/	0x7203,
/*	AND.L	D0,D1	*/	0xC280,
/*	ROR.L	#2,D1	*/	0xE499,
/*	LSR.L	#2,D0	*/	0xE488,
/*	MOVE.L	D2,-(A7) */	0x2F02,
/*	TDIVU.L	D2,D2:D0 */	0x4C42, 2,
/*	DIVU.L	(A7)+,D2:D1 */	0x4C5F, 0x1402
  };

  #undef ifix16div
  #pragma parameter __D0 ifix16div(__D1,__D2)
  extern uint32 ifix16div(uint32, uint32) = {
/*	SWAP	D1	*/	0x4841,
/*	MOVE.L	D1,D0	*/	0x2001,
/*	CLR.W	D1	*/	0x4240,
/*	EXT.L	D1	*/	0x48C1,
/*	DIVS.L	D2,D1:D0 */	0x4C42, 0xC01,
/*	BVC.S	end	*/	0x680C,
/*	EOR.L	D2,D1	*/	0xB581,
/*	SPL	D0	*/	0x5AC0,
/*	EXTB.L	D0	*/	0x49C0,
/*	EORI.L	#$80000000,D0*/	0x0A80, 0x8000, 0
  };
  #pragma parameter __D0 ufix16mul(__D0, __D2)
  extern ufixed ufix16mul(ufixed, ufixed) = {
/*	MULU.L	D2,D2:D0 */	0x4C02, 0x402,
/*	ADD.W	D0,D0	*/	0xD040,
/*	MOVE.W	D2,D0	*/	0x3002,
/*	MOVEQ	#0,D2	*/	0x7400, 
/*	SWAP	D0	*/	0x4840,
/*	ADDX.L	D2,D0	*/	0xD182
  };
 #elif LONG_LONG
  #define ufract32mul(x,y) ((uint32)((uint64)(uint32)(x)*(uint32)(y)>>32))
  #define ufract32div(x,y) (uint32)(((uint64)(uint32)(x)<<32)/(uint32)(y))

  #define ifract30mul(x,y) ((int32)((int64)(int32)(x)*(int32)(y)>>30))
  #define ufract30mul(x,y) ((uint32)((uint64)(uint32)(x)*(uint32)(y)>>30))

  #define ifract30div(x,y) (int32)(((int64)(int32)(x)<<30)/(int32)(y))
  #define ufract30div(x,y) (uint32)(((uint64)(uint32)(x)<<30)/(uint32)(y))

  #define ufix16mul(x,y) (ufixed)((uint64)(ufixed)(x)*(ufixed)(y)+32768>>16)
 #else
  #define ufract32mul(x,y) (uint32)(((double)(uint32)(x)*(uint32)(y))/4294967296.0)
  #define ufract32div(x,y) (uint32)(((double)(uint32)(x)*4294967296.0)/(uint32)(y))

  #define ifract30mul(x,y) (int32)(((double)(int32)(x)*(int32)(y))/1073741824.0)
  #define ufract30mul(x,y) (uint32)(((double)(uint32)(x)*(uint32)(y))/1073741824.0)

  #define ifract30div(x,y) (int32)(((double)(int32)(x)*1073741824.0)/(int32)(y))
  #define ufract30div(x,y) (uint32)(((double)(uint32)(x)*1073741824.0)/(uint32)(y))

  #define ufix16mul(x,y) (ufixed)((ufixed)(x)/65536.0*(ufixed)(y)+.5)
 #endif

ifixed vfix16mul(ifixed,
	ifixed);
ifixed vfix16div(ifixed,
	ifixed);

/* `sqrt(a**2+b**2)` or `sqrt((|a|/|b|)**2+1)*|b|` */
uint32 ihypot2(int32,
	int32);
uint32 uhypot2(uint32,
	uint32);
/* 2011.06.28 */
uint32 ihypot3(int32,
	int32,
	int32);
uint32 uhypot3(uint32,
	uint32,
	uint32);

uint32 usqrt(uint32 n);
uint32 usqrt64(uint64 n);

/*
 All values of [0,4.0) represent [0,2*pi).

   The larger theta is in it's lower 30 bits, the less accurate the result
 (in between the entries in the table), although it is still accurate enough
 for 16 bit samples.
*/
ifract fastsin(ufract theta);
 #define fastcos(t) fastsin((t)+0x40000000u) /* phase shift */

ufract ifractatan2(int32 x,
	int32 y);
ufract ufractatan2(uint32 x,
	uint32 y);
/* if out of range (-1,1) returns min/max values */
ifract fractasin(ifract x);

int cntlz32(uint32 x);

/* built-in instructions */
 #ifdef env68k
  #pragma parameter __D0 _cntlz32(__D0)
  extern int _cntlz32(uint32) = {
/*	BFFFO	D0{0:0},D0 */	0xEDC0, 0
  };
  #define cntlz32(x) _cntlz32(x)
 #elif defined(__MRC__)
  #define cntlz32(x) (int)__cntlzw((uint32)(x))
 #elif defined(__POCC__) && defined(envx86)
  __inline int _cntlz32(uint32 v)
  {
	__asm
	{ /* default -1 for 32 result */
		xor	eax,eax
		dec	eax
		bsr	eax,v 
		neg	eax
		add	eax,31
		mov	v,eax
	}
	return v;
  }
  #define cntlz32(x) _cntlz32(x)
 #endif

 #define __FTRIG_H
#endif
