/*
 ini.h
  interpreter 1.7

 Revised: 2013.04.03
*/

/* ---- constants */

enum
{ /* types - values -1 to -255 are pascal string sizes */
	_int8		=-256,
	_int16		=-257,
	_int32		=-258,
	_int64		=-259,

	_uint8		=-260,
	_uint16		=-261,
	_uint32		=-262,
	_uint64		=-263,
	/* printed in hexadecimal, unclipped */
	_hint8		=-264,
	_hint16		=-265,
	_hint32		=-266,
	_hint64		=-267,

	_ifixed		=-268,	/* i16:16 */
	_ufixed		=-269,	/* u16:16 */
	_ifract		=-270,	/* i2:30 */
	_ufract		=-271,	/* u2:30 */

	_bool		=-272,	/* non-zero byte */
	_flt16		=-273,	/* f16:0 */
	_float		=-274,	/* IEEE 32-bit (s1.e8.i23) */
	_double		=-275,	/* IEEE 64-bit (s1.e11.i52) */

	_ufrc8		=-276,	/* u0:8 */
	_ufrc16		=-277,	/* u0:16 */
	_ufrc32		=-278,	/* u0:32 */
	_pt16		=-279,	/* v/h */

	_rgb48		=-280,	/* u16{r,g,b} */
	_rect16		=-281	/* i16[4] */
};

/* ---- typedefs */

#ifdef LONG_TREC
 typedef long		trec_size;
 typedef unsigned long	trec_offs;
#else
 typedef short		trec_size;
 typedef unsigned short	trec_offs;
#endif

/* text object */
typedef struct tobj
{ /* NULL str = identifier; otherwise, bit #31 = not allocated */
	long		z;
	char		*s;
} tobj;

/* record object */
typedef struct trec
{
	trec_size	z;
	trec_offs	o;
	char		*s;
} trec;

/* normal offset record macro */
#define O_TREC(t,s,n)			{t, (int)&((s*)0)->n, #n}
/* C and Pascal string record macros */
#define C_TREC(s,n)			{sizeof(((s*)0)->n), (int)&((s*)0)->n, #n}
#define P_TREC(s,n)			{1-sizeof(((s*)0)->n), (int)&((s*)0)->n, #n}

/* context structure */
typedef struct tcon
{
	const trec	*d;
	char		*s;
	void		*p;
} tcon;

#define O_TCON(d,s)			{d, #s, &s}

/* ---- prototypes */

int deltobj(tobj *o);

const tcon *gettcon(const tcon *l,
	const tobj *o);
int settrec(const trec *l,
	const tobj *d,
	const tobj *s,
	char *p, /* from the tcon */
	char lit); /* literal ('...') data */
int print_trec(struct FILE *f,
	const trec *l,
	const char *p); /* from the tcon */

long read_ini(struct FILE *f,
	const tcon *cl,
	long m); /* error mask (pass zero for no printed errors) */
long dump_ini(struct FILE *f,
	const tcon *l);
