/*
 interpreter.h
  v1.7.4

  Created: 2010.03.21
  Revised: 2013.05.13

 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

2011.09.17
 nixed array16/32 codes and moved list from xF3 to xF1.  project calls for lean code.
 x4-x7 are now binary.
2012.04.03
 removed stale comments
2012.06.07
 reduced form: 12 packed bytes per item (triple-int32).
 revised types, making all xF4+ binary numbers.
2012.08.05
 couldn't debug the reduced form; back to 16 bytes.
*/

#ifndef __INTERPRETER_H
#include "../enviro64.h"

/* ---- enums */

enum
{ /* type codes: sizes {..12|16, 4|n/a, n/a, 4,4,4,8, 1,2,4,8, 1,2,4,8} */
  /* global items are defined with explicit pointers instead of references */
	_t_str,		/* 0=empty string */
  /* all values in between are explicit string sizes */
	_t_list	= 0x7FFFFFF1, /* list_item (16 bytes full, 12 in stack) */
	_t_func,	/* const, native code void* or func (int32*) */
	_t_binary,	/* const, no specific format: void* */
	_t_ifixed,	/* x4 - signed 16:16 */
	_t_ufixed,	/* unsigned 16:16 */
	_t_float,
	_t_double,
	_t_int8,	/* x8 - bit #3 = integer - size is 1<<(z&3) */
	_t_int16,
	_t_int32,
	_t_int64,
	_t_uint8,	/* xC */
	_t_uint16,
	_t_uint32,
	_t_uint64	/* xF */
};

enum
{ /* error codes */
	_e_param_separation	= 1,	/* white space */
	_e_return		= 0x7FFFFFFF,

	_e_general_error	=-1,	/* syntax/param error */
	_e_memory_error		=-2,	/* alloc failed */
	_e_read_error		=-3,	/* I/O read error */
	_e_write_error		=-4,	/* I/O write error */
	_e_incomplete_error	=-5,	/* expression missing symbols */
	_e_open_interrogator	=-6,	/* incomplete error for '?' */
	_e_no_identifier	=-7,	/* bad/missing ID */
	_e_handler_error	=-8,	/* ID handler not supplied for op */
	_e_stack_error		=-9,	/* too many or missing parentheses */
	_e_range_error		=-10,	/* displacement too large for size */
	_e_open_comment		=-11,	/* comment still open at end */
	_e_open_string		=-12,	/* unterminated string */
	_e_identifier_conflict	=-13,	/* ID already added */
	_e_op_error		=-14,	/* can't encode that! */
	_e_label_error		=-15,	/* can't add this as a label here */
	_e_language_error	=-16,	/* not equipped to process this */
	_e_macro_error		=-17,	/* no such macro */
	_e_cannot_optimize	=-18	/* can't remove/reduce this code */
};

/* ---- constants */

#ifdef mac
 #define CHR_NE				'\255' /* inequal */
 #define CHR_LE				'\262' /* lesser or equal */
 #define CHR_GE				'\263' /* greater or equal */
 #define CHR_RA				'\303' /* radical (square root) */
 #define CHR_NB				'\312' /* non-breaking space */
 #define CHR_MN				'\320' /* minus "en-dash" */
 #define CHR_DV				'\326' /* divide */
 #define CHR_NL				'\015' /* new line: CR */
#else /* ISO-8895-1-compatible */
/* Windows-1252 */
 //no NE (U+2260)
 //no official LE (U+2264)
 //no official GE (U+2265)
 //no RA (U+221A)
 #define CHR_NB				'\240' /* non-breaking space */
 #define CHR_DV				'\367' /* divide */
 //en-dash (U+2013) not in ISO-8859-1
 #define CHR_MN				'\226' /* minus "en-dash" */
 #define CHR_NL				'\012' /* new line: LF */
#endif

#ifndef CHR_MN
 #define CHR_MN				'-'
#endif

#define CASE_NUMERIC			case'0':case'1':case'2':case'3':case'4': \
	case'5':case'6':case'7':case'8':case'9':
#define CASE_ALPHABETIC			case'A':case'B':case'C':case'D':case'E': \
	case'F':case'G':case'H':case'I':case'J':case'K':case'L':case'M':case'N': \
	case'O':case'P':case'Q':case'R':case'S':case'T':case'U':case'V':case'W': \
	case'X':case'Y':case'Z':	case'a':case'b':case'c':case'd':case'e': \
	case'f':case'g':case'h':case'i':case'j':case'k':case'l':case'm':case'n': \
	case'o':case'p':case'q':case'r':case's':case't':case'u':case'v':case'w': \
	case'x':case'y':case'z':
#define CASE_BINARY_OPERATOR		case'%':case'&':case'*':case'/':case'<': \
	case'=':case'>':case'^':case'|':

/* ---- typedefs/structures */

#ifdef env68k
 #define double_t			long double
#else
 #define double_t			double
#endif

#if TARGET_OS_MAC
 #define INFINITY			__inf()
 double_t scalb(double_t,
	long);
#elif __POCC__
 #define scalb				scalbn
#endif

#if EVAL_FLOATING
typedef double		evalvalue;
typedef double		evalparse;
 #define EVAL_HUGEVAL			(evalvalue)((1L<<22)*(evalvalue)(1L<<30)-1)
 #define parse_eval(s,e,i)		parse_number(s,e,i,1)
 #define _t_eval			_t_double
#else
typedef int32		evalvalue;
typedef long int	evalparse;
 #define EVAL_HUGEVAL			((evalvalue)1<<(sizeof(evalvalue)<<3)-1)
 #define parse_eval(s,e,i)		readanyinteger(s,e-s,i)
 #define _t_eval			_t_int32
#endif

typedef union item_data/*(8)*/
{
	evalvalue	val, *vptr;
	char		str[8], *sptr;
	uint8		raw[8], *rptr;
	void		*ptr;

	struct list_item	*lptr;
} item_data;

/* non-reference pointers are mallocs when >=8 bytes */
#define ITEM_HAS_MALLOC(l)		((uint32)(l)->type-9<_t_ifixed-9)
#define ITEM_IS_REF(l)			((l)->type&0x80000000U)
#define ITEM_IS_RAWSTR(l)		((uint32)(l)->type<9)
#define ITEM_IS_RAWVAL(l)		((l)->type>=_t_ifixed)
#define ITEM_IS_MALLOC_REF(l)		((l)->type<(_t_ifixed|-1<<31))
#define RAW_VAL_SIZE(t)			(1u<<((t)&3))
//#define ITEM_GET_PTR(l) (ITEM_IS_RAWSTR(l)||ITEM_IS_RAWVAL(l)?(l)->data.raw:(l)->data.ptr)

typedef struct list_item/*(16)*/
{ /* offset is 'type' size for string references */
	int32		index; /* user data */
	int32		type; /* see the codes above; bit 31 = reference */
	item_data	data;
} list_item;

/* item #0 defines the terms to the list; pos is count (>0) */
/* names strlist: hash(first char in low 6 bits), Pascal string */
typedef list_item	*list_t;

/* ---- prototypes */

/* system standard */

int strcasecmp(const char *a,
	const char *b);
int strncasecmp(const char *a,
	const char *b,
	size_t z);

/* extended standard */

/* parses ASCII-compatible chars to return length of an entire line; 2011.08.13 */
/* escape char can be anything, such as '\', or 0 (null) for no escape char */
long sreadline(const char *p,
	long length,
	char escape);
/* same as sallocline, but uses fgetc/ungetc, allocates string w/ reallocfunc */
/* output is capped to 0x7FFFFFF0 bytes */
char *freadline(struct FILE*,
	uint32 *length, /* in/out; can be NULL (C-string result) */
	char *store, /* in-place realloc'd data; can be NULL */
	void *(reallocfunc)(void*, /* can be NULL (default: realloc) */
		size_t),
	uint32 *lecount, /* line-ending count, in/out; can be NULL */
	int escape); /* can be 0 (no escape char) */

int makebasestring(void *outstring,
	const char *basechars, /* can be NULL (use default chars) */
	long value,
	unsigned base,
	int minlength,
	char sign); /* pass zero to interpret unsigned */

/* returns ptr to next character */
const char *readanyinteger(const char *str,
	long length,
	long int *outvalue);
/* added 2013.04.01 */
const char *readanyint64(const char *s,
	const char *e,
	int64 *v);

const char *parse_number(const char *s,
	const char *e,
	double *v,
	uint8 r); /* real (take the fractional point) */
double_t scal10(double_t v,
	int e);
/* this requires that the characters have already been processed */
int parse_literal(const char *s,
	int64 *r,
	long z);

const char *parse_eval_literal(const char *str,
	const char *e,
	evalvalue *v);

/* returns next position (after closing quote) */
const char *parse_textloop(const char *s,
	const char *e,
	long *z, /* in max./out length; negative on memory error */
	uint32 *t, /* output type (Pascal, C or raw); can be NULL */
	char **r, /* will use realloc to buffer this; can be NULL */
	void *u, /* ptr passed to b func */
	void (*b)(void *u,
		char v));

	/* prints a string to file output, including escapes for quote chars */
/* the bounding quote chars are not printed by this function */
/* returns total printed char count */
uint32 print_tchars(struct FILE *f,
	const char *s,
	long z,
	char q);

const char *parse_wordbreak(const char *s, /* does nothing if either are NULL */
	const char *e,
	uint8 *h); /* non-zero first char hash (can be NULL); 2011.09.15 */
const char *parse_rwordbreak(const char *s,
	const char *f);
const char *parse_skipwsp(const char *s,
	const char *e);
const char *parse_rskipwsp(const char *s,
	const char *f);

/* each division by zero will yield EVAL_HUGEVAL */
const char *parse_parameter(const char *s,
	long length,
	char *outstring, /* can be NULL; otherwise, expects at least 24 chars allocated */
	int *outerr,
	const char *(*getv)(const char *s,
		const char *e,
		evalvalue *i,
		void *u),
	const char *(*setv)(const char *s,
		const char *e,
		evalvalue i,
		void *u),
	void *userdata, /* for getv */
	evalvalue (*evalindir)(evalvalue)); /* indirection; (`@`); can be NULL */

/* returns negative on error, otherwise length till end (comment, line ending) */
long evaluate_line(const char *s,
	const char *e,
	evalvalue *v,
	int *err,
#ifdef ALLOW_LONG_COMMENTS
	uint8 *com, /* within a long-comment (` #[..]# `)*/
#endif
	const char *(*getv)( /* can be NULL */
		const char *s,
		const char *e,
		evalvalue*,
		void *u),
	const char *(*setv)( /* can be NULL */
		const char *s,
		const char *e,
		evalvalue,
		void *u),
	void *u,
	evalvalue (*evalindir)(evalvalue)); /* can be NULL */

/* ---- lists */

/* if successful, returns the ptr to the new entry (beyond str item #0) */
list_item *list_add(list_t *p, /* can be NULL (new list) - item #1 still returned */
	long x,
	long z,
	void *v); /* data only copied in raw case */
void list_dispose(list_t p);

long item_getsize(const list_item *l);

void *item_resize(list_item *l,
	long z);
long item_set(list_item *l,
	long x,
	long z,
	void *p,
	int r); /* reserve raw types (convert instead of change type) */
double item_to_double(long z,
	const void *p);

uint32 getnamesitemindex(const uint8 *s,
	int32 sz,
	const char *n,
	uint8 nz,
	uint8 h); /* hash char from parse_wordbreak */

#define __INTERPRETER_H
#endif
