/*
 x6502.h
  used by the 2A0x functions

 * MacFCEU 0.7
 *
 * Copyright notice for this file:
 *  x6502.h, x6502struct.h Copyright (C) 2002 Xodnizel
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef _X6502H

/* ---- constants */

#define NTSC_CPU		1789772.7272727272727272	/* 19687500 11ths */
#define PAL_CPU			1662607.125			/* 1.6xMHz PAL */

#define N_FLAG			0x80	/* Negative */
#define V_FLAG			0x40	/* oVerflow */
#define U_FLAG			0x20	/* Unused status flag (always 1) */
#define B_FLAG			0x10	/* Break */
#define D_FLAG			0x08	/* Decimal */
#define I_FLAG			0x04	/* Interrupt disable */
#define Z_FLAG			0x02	/* Zero */
#define C_FLAG			0x01	/* Carry */

#define FCEU_IQEXT		0x001
#define FCEU_IQFDS		0x002	/* prev. IQEXT2 */
/* ... */
#define FCEU_IQRESET		0x020
#define FCEU_IQNMI2		0x040	/* Delayed NMI, gets converted to *_IQNMI */
#define FCEU_IQNMI		0x080
#define FCEU_IQDMC		0x100
#define FCEU_IQFCOUNT		0x200
#define FCEU_IQTEMP		0x800

typedef struct X6502u32 /* for immediate 32-bit extension */
{
	uint32		PC;
	uint32		A;
	uint32		X;
	uint8		Y, S, mooPI, P;
	uint32		IA;		/* last internal 16 bit interrupt read address */
	uint32		jammed;
} X6502u32;

typedef struct X6502struct/*(36)*/
{
#if BIG_ENDIAN
	uint16		PCzero, PC;
	uint8		Azero[3], A;
	uint8		Xzero[3], X;
	uint8		Y, S, mooPI, P;
	uint8		IAz0, IAz1, IAx2, IAx3;
#else
	uint16		PC, PCzero;
	uint8		A, Azero[3];
	uint8		X, Xzero[3];
	uint8		Y, S, mooPI, P;
	uint8		IAx3, IAx2, IAz1, IAz0;
#endif
	uint8		zero[3], jammed;
	int32		tcount;		/* Temporary cycle counter. */
	int32		count;
	uint32		IRQlow;		/* Simulated IRQ pin held low, bits emulator-specific. */
} X6502struct, X6502;

typedef union X6502U
{
	X6502		low;
	X6502u32	u32;
} X6502U;

/* ---- globals */

extern uint32 timestamp;
extern X6502U X6502F;
#define XFrame X6502F.low

extern void FP_FASTAPASS(1) (*MapIRQHook)(
	int a); /* instruction cycle count */

extern void FP_FASTAPASS(1) (*DebugStepHook)(
	int c); /* count (used by the debugger - pass zero) */

extern void FP_FASTAPASS(1) (*X6502_Run)(
	int32 cycles); /* PPU cycle count */

/* ---- prototypes */

/* positive to OR, eitherwise AND (~x) */
void X6502_DebugLevel(int bits); /* 1 is user, 2 is cheat.c, 4 is GameGenie */

void X6502_Init(void);
void X6502_Reset(void);
void X6502_Power(void);

uint8 FASTAPASS(1) X6502_DMR(uint32 A);
void FASTAPASS(2) X6502_DMW(uint32 A, uint8 V);

void FASTAPASS(1) X6502_IRQBegin(int w);
void FASTAPASS(1) X6502_IRQEnd(int w);

#define _X6502H
#endif
