/*
 zlib/contrib/minizip/miniunz.c
  Version 1.01e+, February 12th, 2005

   Copyright (C) 1998-2005 Gilles Vollant

Mods by Adam P.
Fixed poor gramar as well.
*/

#define LAST_REVISION_DATE      "2013.03.19 00:44 EDT"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <errno.h>
#include <fcntl.h>

#ifdef mac
# include ":::zlib.h"
# include "::macos:macos7-9.h"
# include <Files.h>
/* so the user can cancel */
# define signalbreak()  ((*((volatile UInt8*)0x179)&*((volatile UInt8*)0x17A))!=0)
# define TPATH_SEP      ':'
# define TPATH_EXTRA    1
# define MAXNAMELENGTH  31
#else
# define signalbreak()  0
# define TPATH_SEP      '/'
# define TPATH_EXTRA    0
# define MAXNAMELENGTH  255

# ifdef unix
#  include <unistd.h>
#  include <utime.h>
# else
#  include <direct.h>
#  include <io.h>
# endif
#endif

#include "unzip.h"

#ifndef local
# define local static
#endif


#define CASESENSITIVITY (0)
#define WRITEBUFFERSIZE (8192)
#define MAXPATHLENGTH   (256)

#ifdef WIN32
# define USEWIN32IOAPI
# include "iowin32.h"
#endif
/*
  mini unzip, demo of unzip package

  usage :
  Usage : miniunz [-exvlo] file.zip [file_to_extract] [-d extractdir]

  list the file in the zipfile, and print the content of FILE_ID.ZIP or README.TXT
    if either exists
*/

int platformfpath           OF((char*,char*));
char *printablefpath        OF((char*));

void setfileinfo            OF((const char*,const unz_file_info*,const char*));

int makedir                 OF((const char*));


#ifdef mac
static void backwritehash(unsigned char *d,
    unsigned a,
    int z,
    unsigned b,
    const char *t)
{
    unsigned       r;

    do
        r = a, a/= b, r-= a*b, *--d = t[r];
    while(--z!=0);
    *--d = '-';
}

static void *shortenfpath(void *name,
    void *point,
    unsigned a)
{
    unsigned char      *d, *f = NULL, *e, c;

    if ((d=(unsigned char*)point)==NULL)
        return NULL;
    if (name==NULL && d<=(unsigned char*)name)
        return d;
    /* scan to end of name, finish hash */
    while((c=*d++)!=0 && (c+=-'/')!=0 && (c+='/'-'\\')!=0 && (c+='\\'-':')!=0)
        a = (a<<5)^c;
    e = --d;

    /* scan beginning of name, gather dot-extension */
    do
    {
        if ((c=*--d)==':')
            {++d; break;}
        if (f!=NULL)
            continue;
        if (c=='.' && e-(f=d)>8)
        { /* truncate extension to 8, including the '.' */
            memmove(8+d,e,1+strlen((char*)e));
            e = 8+d;
        }
    } while(d>name);

    if (f==NULL)
        f = e;
    d+= MAXNAMELENGTH+f-e;
    e+= d-f;
    memmove(d,f,1+strlen((char*)f));

    backwritehash(d,a,6,40,"0123456789abcdefghijklmnopqrstuvwxyz_~! ");

    return e;
}
#endif

/* returns -1 if NULL, 0 if unchanged, bit 0 set if changed, bit 1 set if shortened */
int platformfpath(char *n,char*m)
{
    unsigned char      *d, c;
    int                 r = 0;
(void)c;
    *m = 0;
    if ((d=(unsigned char*)n)==NULL || !*d)
        return -1; /* no path */
#ifdef mac
    if (!memcmp("__MACOSX/",d+1,9))
    {
        char    *s;

        strcpy(n+1,n+10);
        if ((s=strstr(n,"._"))==NULL)
            r =-2; /* bad path */
        else
            strcpy(s,s+2);
        *m = 1;
        /* rsrc established; remove indicators */
    }
    if ((c=*d++)!=0)
    {
        unsigned        z = 0, a = 0; /* name size, hash for unique renaming */

        do
            if ((c+=-'/')==0 || (c+='/'-'\\')==0)
                d[-1] = ':', z = 0, a = 0, r|= 1;
            else if (a=(a<<5)^c,++z>MAXNAMELENGTH)
                d = shortenfpath(n,d,a), r|= 2;
        while((c=*d++)!=0);
    }
#endif
    return r;
}

/* Mac OS X header
"\x00\x05\x16\x07\x00\x02\x00\x00\x4D\x61\x63\x20\x4F\x53\x20\x58"
"\x20\x20\x20\x20\x20\x20\x20\x20\x00\x02\x00\x00\x00\x09\x00\x00"
"\x00\x32\x00\x00\x00\x20\x00\x00\x00\x02\x00\x00\x00\x52\x00\x00";
*/

/* any char outside [0x20,0xFF) -> space */
char *printablefpath(char *name)
{
    char            *d;
    unsigned char   c;

    if ((d=name)==NULL)
        return NULL;
    if ((c=*d++)!=0)
        do
            if ((c+=-0x20)>=0xFF-0x20)
                d[-1] = ' ';
        while((c=*d++)!=0);

    return name;
}


void setfileinfo(const char *filepath,
	const unz_file_info *file_info,
	const char *tc)
{
#ifdef WIN32
    HANDLE      hFile;
    FILETIME    ftm,ftLocal,ftCreate,ftLastAcc,ftLastWrite;

    hFile = CreateFile(filepath,GENERIC_READ | GENERIC_WRITE,
        0,NULL,OPEN_EXISTING,0,NULL);
    GetFileTime(hFile,&ftCreate,&ftLastAcc,&ftLastWrite);
    DosDateTimeToFileTime((WORD)(file_info->dosDate>>16),(WORD)file_info->dosDate,&ftLocal);
    LocalFileTimeToFileTime(&ftLocal,&ftm);
    SetFileTime(hFile,&ftm,&ftLastAcc,&ftm);
    CloseHandle(hFile);
#else
    struct tm newdate;
    tm_unz tmu_date = file_info->tmu_date;

    newdate.tm_sec = tmu_date.tm_sec;
    newdate.tm_min = tmu_date.tm_min;
    newdate.tm_hour = tmu_date.tm_hour;
    newdate.tm_mday = tmu_date.tm_mday;
    newdate.tm_mon = tmu_date.tm_mon;
    if (tmu_date.tm_year > 1900)
        newdate.tm_year = tmu_date.tm_year - 1900;
    else
        newdate.tm_year = tmu_date.tm_year;
    newdate.tm_isdst = -1;

    {
        time_t          ziptime = mktime(&newdate);
#ifdef mac
        CInfoPBRec      cinfo;
        unsigned char   name[256];
        int             len = strlen(filepath);

        if (len>255)
            return;
        memcpy(1+name,filepath,len);
        name[0] = len;
        cinfo.hFileInfo.ioCompletion = NULL;
        cinfo.hFileInfo.ioFDirIndex = 0;
        cinfo.hFileInfo.ioFlAttrib = 0;
        cinfo.hFileInfo.ioACUser = 0;
        cinfo.hFileInfo.ioNamePtr = name;
        cinfo.hFileInfo.ioVRefNum = vRefNumLocal;
        cinfo.hFileInfo.ioDirID = cinfo.hFileInfo.ioFlParID = 0;
        if (PBGetCatInfoSync(&cinfo)!=noErr)
            return;
        cinfo.hFileInfo.ioFlMdDat = ziptime;
        /* only override creation date if new; 2009.11.26 17:46 EST */
        if (cinfo.hFileInfo.ioFlCrDat==0
                || cinfo.hFileInfo.ioFlCrDat>cinfo.hFileInfo.ioFlMdDat)
            cinfo.hFileInfo.ioFlCrDat = cinfo.hFileInfo.ioFlMdDat;

        /* text!; 2009.07.11 12:32 EST */
        if (file_info->internal_fa&1)
            cinfo.hFileInfo.ioFlFndrInfo.fdType = 'TEXT';
	/* Windows bit #1, read-only; bit #14, executable icon - 2012.03.02 */
        cinfo.hFileInfo.ioFlAttrib|= file_info->external_fa&kioFlAttribLockedMask;

        suffixtofileandcreator(1+name,len,&cinfo.hFileInfo.ioFlFndrInfo.fdType);
        cinfo.hFileInfo.ioDirID = cinfo.hFileInfo.ioFlParID = 0;
        if (tc)
            memcpy(&cinfo.hFileInfo.ioFlFndrInfo.fdType,tc,8);
        PBSetCatInfoSync(&cinfo);
#elif defined(unix)
        struct utimbuf ut;

        ut.actime = ut.modtime = ziptime;
        utime(filepath,&ut);
#endif
    }
#endif
}

#if defined(WIN32) || !defined(unix)
# define mymkdir(x)     mkdir(x)
#else
# define mymkdir(x)     mkdir(x,0775)
#endif

int makedir(newdir)
    const char *newdir;
{
    unsigned char      *p, hold;
    char               *buffer;
    int                 len;

    if ((len=(int)strlen(newdir))<=0 || (buffer=(char*)malloc(len+1))==NULL)
        return 0;

    memcpy(buffer,newdir,len);
    buffer[len] = '\0';

    if (buffer[len-1]==TPATH_SEP)
        buffer[len-1] = '\0';

    if (mymkdir(buffer)==0)
    {
        free(buffer);
        return 1;
    }

    p = (unsigned char*)(1+buffer);

    do
    {
        while((hold=*p)!=0 && hold!='\\' && hold!=TPATH_SEP)
            p++;
        *p = 0;
        if (mymkdir(buffer)==-1 && errno==ENOENT)
        {
            fprintf(stderr,"couldn't create directory '%s'\n",
                buffer);
            free(buffer);
            return 0;
        }
    } while((*p++=hold)!=0);

    free(buffer);
    return 1;
}

local void do_banner()
{
    printf(
"MiniUnz 1.01b+, demo of zLib + Unz package written by Gilles Vollant\n"
"more info at http://www.winimage.com/zLibDll/unzip.html\n"
" - src Mod-Date " LAST_REVISION_DATE "\n\n"
        );
}

local void do_help()
{
    printf(
"Usage : miniunz [options] file.zip [file_to_extr.] [-d extractdir]\n"
"\n"
"  -e  Extract without pathname (junk paths)\n"
"  -x  Extract with pathname\n"
"  -v  list files\n"
"  -l  list files\n"
"  -d  directory to extract into\n"
"  -o  overwrite files without prompting\n"
"  -p  extract crypted file using password\n\n");
}


local int do_list(uf,vb)
    unzFile uf;
    int vb;
{
    unz_global_info     gi;
    uLong               i;
    uLong               n = 0, total_compressed = 0, total_uncompressed = 0;
    int                 err;

    if ((err=unzGetGlobalInfo(uf,&gi))!=UNZ_OK)
        fprintf(stderr,"error %d with zipfile in unzGetGlobalInfo.\n",
            err);

    if (vb<=1)
        printf(
"  Length    Method    Size   Ratio    Date     Time   CRC-32    Name\n"
"----------  ------  -------- ----- ---------- -----  --------   ----\n"
            );
    else
        printf(
"  Length   Method   Size      Date      Time    CRC-32    Vers    Info I/E Attrs.   Name\n"
"---------- ------ -------- ---------- -------- -------- -------- -------- --------  ----\n"
            );

    for(i=0; i<gi.number_entry; ++i)
    {
        char            filename_inzip[256];
        unz_file_info   file_info;
        uLong           ratio = 0;
 
        /* ["Unkn. " - by default to make sure it's set]; 2009.07.11 04:53 EST */
        const char     *string_method = "??? ";
        char            charCrypt = ' ';

        if (signalbreak())
            break;
        if ((err=unzGetCurrentFileInfo(uf,&file_info,filename_inzip,
                sizeof(filename_inzip),NULL,0,NULL,0)) != UNZ_OK)
        {
            fprintf(stderr,"error %d with zipfile in unzGetCurrentFileInfo\n",
                err);
            break;
        }
        if (file_info.uncompressed_size>0)
        {
            ++n;
            total_compressed+= file_info.compressed_size;
            total_uncompressed+= file_info.uncompressed_size;
            ratio = (file_info.compressed_size*100)/file_info.uncompressed_size;
        }
        /* display a '*' if the file is crypted */
        if (file_info.flag&1)
            charCrypt='*';

        if (file_info.compression_method==0)
            string_method="Raw "; /*"Stored"*/
        else if (file_info.compression_method==Z_DEFLATED)
            switch((uInt)(file_info.flag & 0x6)/2)
            {
            default: string_method="Dfl:N"; break;
            case 1: string_method="Dfl:X"; break;
            case 2: string_method="Dfl:f"; break;
            case 3: string_method="Dfl:F"; break;
            }

        if (vb<=1)
            printf(
"%10lu  %6s%c%9lu %3lu%%  %4.2lu.%2.2lu.%2.2lu %2.2lu:%2.2lu  %8.8lX   %s\n",
                    file_info.uncompressed_size,string_method,
                    charCrypt,
                    file_info.compressed_size,
                    ratio,
                    (uLong)file_info.tmu_date.tm_year,
                    (uLong)file_info.tmu_date.tm_mon + 1,
                    (uLong)file_info.tmu_date.tm_mday,
                    (uLong)file_info.tmu_date.tm_hour,
                    (uLong)file_info.tmu_date.tm_min,
                    (uLong)file_info.crc,
                    filename_inzip);
        else
            printf(
"%10lu %5s%c%9lu %4.2lu.%2.2lu.%2.2lu %2.2lu:%2.2lu:%2.2lu %8.8lX %8.8lX %8.8lX %8.8lX  %s\n",
                    file_info.uncompressed_size,string_method,
                    charCrypt,
                    file_info.compressed_size,
                    (uLong)file_info.tmu_date.tm_year,
                    (uLong)file_info.tmu_date.tm_mon + 1,
                    (uLong)file_info.tmu_date.tm_mday,
                    (uLong)file_info.tmu_date.tm_hour,
                    (uLong)file_info.tmu_date.tm_min,
                    (uLong)file_info.tmu_date.tm_sec,
                    (uLong)file_info.crc,
                    ((uLong)file_info.version<<16)|(uLong)file_info.version_needed,
                    (((uLong)file_info.flag&-8)<<16)|(uLong)file_info.internal_fa,
                    (uLong)file_info.external_fa,
                    filename_inzip);

        if ((i+1)<gi.number_entry)
            if ((err=unzGoToNextFile(uf))!=UNZ_OK)
            {
                fprintf(stderr,"error %d with zipfile in unzGoToNextFile\n",
                    err);
                break;
            }
    }

    if (n>((vb<=1) ? 1:0)) /* no point in displaying totals/ratios more than once */
        printf(
"%10lu         %*lu %3lu%%  Total\n",
                total_uncompressed,
                ((vb<=1) ? 9:7),total_compressed,
                (uLong)((total_compressed*100.0)/total_uncompressed));

    return 0;
}


local int do_extract_currentfile(uf,popt_extract_without_path,popt_overwrite,password)
    unzFile uf;
    const int *popt_extract_without_path;
    int *popt_overwrite;
    const char *password;
{
    char                filename[TPATH_EXTRA + MAXPATHLENGTH], tc[8];
    unz_file_info       file_info;
    char               *filename_withoutpath, rsrc;
    int                 rtv = UNZ_OK;
    FILE               *fout = NULL;
    void               *buf;
    uInt                size_buf;
    int                 nativepath_result;
    uLong               ratio = 0;

    if ((rtv=unzGetCurrentFileInfo(uf,&file_info,TPATH_EXTRA + filename,
            MAXPATHLENGTH,NULL,0,NULL,0))!=UNZ_OK)
    {
        fprintf(stderr,"error %d with zipfile in unzGetCurrentFileInfo\n",
            rtv);
        return rtv;
    }

    size_buf = WRITEBUFFERSIZE;
    if ((buf=(void*)malloc(size_buf))==NULL)
    {
        fprintf(stderr,"Error allocating memory\n");
        return UNZ_INTERNALERROR;
    }

#if TPATH_EXTRA
    *filename = TPATH_SEP;
#endif

    {
        unsigned char  *p = (unsigned char*)(filename_withoutpath=filename);
        unsigned char   c;

        if ((c=*p++)!='\0')
            do
                if (c=='/' || c=='\\')
                    filename_withoutpath = (char*)p;
            while((c=*p++)!='\0');
    }

    if (strstr(filename,".DS_Store"))
    {
        rtv = UNZ_OK; goto end;
    }
    if ((nativepath_result=platformfpath(filename,&rsrc))<0)
    {
        if (*filename_withoutpath)
            printf("bad modern path '%s'\n",
                filename);
        rtv = UNZ_OK; goto end;
    }
    if (*filename_withoutpath=='\0')
    {
        if (*popt_extract_without_path==0)
        {
            if (!rsrc)
                printf("creating directory '%s'\n",
                    filename);
            mymkdir(filename);
        }
    } else
    {
        const char     *write_filename;
        int             skip = 0;

        if (*popt_extract_without_path==0)
            write_filename = filename;
        else
            write_filename = filename_withoutpath;

        if ((rtv=unzOpenCurrentFilePassword(uf,password))!=UNZ_OK)
            fprintf(stderr,"error %d with zipfile in unzOpenCurrentFilePassword\n",
                rtv);

        if (*popt_overwrite==0 && rtv==UNZ_OK && !rsrc)
        {
            char        rep=0;
            FILE        *ftestexist;

            if ((ftestexist=fopen(write_filename,"rb"))!=NULL)
            {
                fclose(ftestexist);
                do
                {
                    char        answer[8];

                    /* for it to work in MPW; 2009.07.10 20:27:08 EST */
                    printf("The file '%s' exists. Overwrite ? [y]es, [n]o, [A]ll:\n",
                        write_filename);

                    if (fgets(answer,sizeof(answer)-1,stdin)==NULL || answer[0]==0)
                        exit(EXIT_FAILURE);

                    rep = *answer;
                    if (rep>='a' && rep<='z')
                        rep-= 0x20;
                } while(rep!='Y' && rep!='N' && rep!='A');
            }

            if (rep=='N')
                skip = 1;

            if (rep=='A')
                *popt_overwrite = 1;
        }

        if (!skip && rtv==UNZ_OK)
        {
            /* some zipfiles don't contain directory alone before file */
            if (rsrc)
                fout = fdopen(open(write_filename,O_RDWR|O_BINARY|O_RSRC),"wb");
            else
                fout = fopen(write_filename,"wb");
            if (fout==NULL && *popt_extract_without_path==0
                    && filename_withoutpath!=(char*)filename)
            {
                char    c = filename_withoutpath[-1];

                filename_withoutpath[-1] = '\0';
                makedir(write_filename);
                filename_withoutpath[-1] = c;
                fout = fopen(write_filename,"wb");
            }

            if (!fout)
                fprintf(stderr,errno==EACCES ? "error: '%s' appears to be read-only\n":"error opening '%s'\n",
                    write_filename);
        }

        if (fout)
        {
            printf(" %s '%s'\n",
                rsrc ? "extracting rsrc":nativepath_result>1 ? "extracting (as)":"extracting",write_filename);

            if (rsrc && (rtv=unzReadCurrentFile(uf,buf,82))<0)
                goto errc; /* skip header */
            do
                if ((rtv=unzReadCurrentFile(uf,buf,size_buf))<0)
                {
                errc:
                    fprintf(stderr,"error %d with zipfile in unzReadCurrentFile\n",
                        rtv);
                    break;
                } else if (rtv>0)
                {
                    if (rsrc==1) /* get type/creator from rsrc header */
                        {memcpy(tc,(char*)buf+82,8), rsrc = 2;}
                    if (fwrite(buf,rtv,1,fout)!=1)
                    {
                        fprintf(stderr,"error in writing extracted file\n");
                        rtv = UNZ_ERRNO;
                        break;
                    }
                }
            while(rtv>0);

            if (fout)
                fclose(fout);

            if (rtv==0)
                setfileinfo(write_filename,&file_info,rsrc==2 ? tc:NULL);
        }

        if (rtv==UNZ_OK)
        {
            if ((rtv=unzCloseCurrentFile(uf))!=UNZ_OK)
                fprintf(stderr,"error %d with zipfile in unzCloseCurrentFile\n",
                    rtv);
        } else
            unzCloseCurrentFile(uf); /* don't lose the error */
    }
end:
    free(buf);
    return rtv;
}


local int do_extract(uf,opt_extract_without_path,opt_overwrite,password)
    unzFile uf;
    int opt_extract_without_path;
    int opt_overwrite;
    const char *password;
{
    uLong               i;
    unz_global_info     gi;
    FILE               *fout = NULL;
    int                 err;

    if ((err=unzGetGlobalInfo(uf,&gi))!=UNZ_OK)
        fprintf(stderr,"error %d with zipfile in unzGetGlobalInfo \n",
            err);

    for(i=0; i<gi.number_entry; ++i)
    {
        if (i&4) /* some manual screen update; 2010.03.03 15:21 EST */
            fflush(stdout);
        if (signalbreak()) /* if user wants to stop; 2010.04.10 15:16 EDT */
            {err = -128; break;}
        if (do_extract_currentfile(uf,&opt_extract_without_path,
                &opt_overwrite,
                password)!=UNZ_OK)
            break;

        if ((i+1)<gi.number_entry)
            if ((err=unzGoToNextFile(uf))!=UNZ_OK)
            {
                fprintf(stderr,"error %d with zipfile in unzGoToNextFile\n",
                    err);
                break;
            }
    }

    return 0;
}

local int do_extract_onefile(uf,filepath,opt_extract_without_path,opt_overwrite,password)
    unzFile uf;
    const char *filepath;
    int opt_extract_without_path;
    int opt_overwrite;
    const char *password;
{
    int                 err = UNZ_OK;

    if (unzLocateFile(uf,filepath,CASESENSITIVITY)!=UNZ_OK)
    {
        printf("file '%s' not found in the zipfile\n",
            filepath);
        return 2;
    }

    if (do_extract_currentfile(uf,&opt_extract_without_path,
            &opt_overwrite,password)==UNZ_OK)
        return 0;
    else
        return 1;
}


int main(argc,argv)
    int argc;
    char *argv[];
{
    char                filename_try[MAXPATHLENGTH+16];
    const char         *zipfilename = NULL;
    const char         *filename_to_extract = NULL;
    const char         *password = NULL;
    int                 i;
    int                 opt_do_list = 0;
    int                 opt_do_extract = 1;
    int                 opt_do_extract_withoutpath = 0;
    int                 opt_overwrite = 0;
    int                 opt_extractdir = 0;
    const char         *dirname = NULL;
    unzFile             uf = NULL;

    *filename_try = '\0';

    do_banner();

    if (argc==1)
    {
        do_help();
        return 0;
    } else
        for(i=1; i<argc; ++i)
            if ((*argv[i])=='-')
            {
                const unsigned char     *p = (const unsigned char*)(1+argv[i]);
                unsigned char           c;

                while((c=*p++)!='\0')
                {
                    if (c=='l' || c=='L')
                        opt_do_list = 1;
                    if (c=='v' || c=='V')
                        opt_do_list = 2;
                    if (c=='x' || c=='X')
                        opt_do_extract = 1;
                    if (c=='e' || c=='E')
                        opt_do_extract = opt_do_extract_withoutpath = 1;
                    if (c=='o' || c=='O')
                        opt_overwrite = 1;
                    if ((c=='d' || c=='D') && i+1<argc) /* check; 2010.03.03 15:16 EST */
                    {
                        opt_extractdir = 1;
                        dirname = argv[i+1];
                    }

                    if ((c=='p' || c=='P') && i+1<argc)
                        password = argv[++i];
                }
            } else if (zipfilename==NULL)
                zipfilename = argv[i];
            else if (filename_to_extract==NULL && !opt_extractdir)
                filename_to_extract = argv[i];

    if (zipfilename!=NULL)
    {
#     ifdef USEWIN32IOAPI
        zlib_filefunc_def   ffunc;
#     endif

        strncpy(filename_try,zipfilename,MAXPATHLENGTH-1);
        /* strncpy doesn't append the trailing NULL if the string is too long. */
        filename_try[ MAXPATHLENGTH ] = '\0';

#     ifdef USEWIN32IOAPI
        fill_win32_filefunc(&ffunc);
        uf = unzOpen2(zipfilename,&ffunc);
#     else
        uf = unzOpen(zipfilename);
#     endif
        if (uf==NULL)
        {
            strcat(filename_try,".zip");
#         ifdef USEWIN32IOAPI
            uf = unzOpen2(filename_try,&ffunc);
#         else
            uf = unzOpen(filename_try);
#         endif
        }
    }

    if (uf==NULL)
    {
        fprintf(stderr,"Cannot open '%s'[.zip]\n",
            zipfilename,zipfilename);
        return 1;
    }

    printf("'%s' opened\n",
        filename_try);
    fflush(stdout);

    i = 0;

    if (opt_do_list>0)
        i = do_list(uf,opt_do_list);
    else if (opt_do_extract==1)
    {
#     ifndef mac
        if (opt_extractdir && chdir(dirname))
        {
            fprintf(stderr,"Error changing into '%s', aborting\n",
                dirname);
            exit(-1);
        }
#     endif
        if (filename_to_extract==NULL)
            i = do_extract(uf,opt_do_extract_withoutpath,opt_overwrite,password);
        else
            i = do_extract_onefile(uf,filename_to_extract,
                opt_do_extract_withoutpath,opt_overwrite,password);
    }

    unzCloseCurrentFile(uf);

    printf("\n'%s' closed.\n\n",
        filename_try);

    return i;
}
